#!/usr/bin/env python
from __future__ import print_function
import shutil
import logging
import unittest
from ruffus.proxy_logger import make_shared_logger_and_proxy, setup_std_shared_logger
from ruffus import originate, transform, suffix, merge, pipeline_run, Pipeline
import sys

"""

    test_with_logger.py

"""


import os
tempdir = os.path.relpath(os.path.abspath(os.path.splitext(__file__)[0])) + "/"
input_file_names = [os.path.join(tempdir, "%d.1" % fn) for fn in range(20)]
final_file_name = os.path.join(tempdir, "final.result")
try:
    os.makedirs(tempdir)
except:
    pass


# add grandparent to search path for testing
grandparent_dir = os.path.abspath(
    os.path.join(os.path.dirname(__file__), "..", ".."))
sys.path.insert(0, grandparent_dir)

# module name = script name without extension
module_name = os.path.splitext(os.path.basename(__file__))[0]


# funky code to import by file name
# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#   imports


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#   Tasks


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

def write_input_output_filenames_to_output(infiles, outfile, logger_proxy, logging_mutex):
    """
    Helper function: Writes input output file names and input contents to outfile
    """
    with open(outfile, "w") as oo:
        # save file name strings before we turn infiles into a list
        fn_str = "%s -> %s" % (infiles, outfile)

        # None = []
        if infiles is None:
            infiles = []
        # str = [str]
        if not isinstance(infiles, list):
            infiles = [infiles]

        max_white_space = -2
        # write content of infiles indented
        for infile in infiles:
            with open(infile) as ii:
                for line in ii:
                    oo.write(line)
                    max_white_space = max(
                        [max_white_space, len(line) - len(line.lstrip())])
        # add extra spaces before filenames
        oo.write(" " * (max_white_space + 2) + fn_str + "\n")

    with logging_mutex:
        logger_proxy.info(fn_str)


#
#   Make logger
#
#import logging
args = dict()
args["file_name"] = os.path.join(tempdir, module_name + ".log")
args["level"] = logging.DEBUG
args["rotating"] = True
args["maxBytes"] = 20000
args["backupCount"] = 10
args["formatter"] = "%(asctime)s - %(name)s - %(levelname)6s - %(message)s"

if sys.version_info[0] == 3 and sys.version_info[1] == 2 and __name__ != "__main__":
    print(
        """
    888888888888888888888888888888888888888888888888888888888888888888888888888

        ERROR:

    This unit test can not be run as a python module (python -m unittest xxx)
    due to the interaction of bugs / misfeatures in the multiprocessing module
    and python3.2

        See http://bugs.python.org/issue15914
            http://bugs.python.org/issue9573

    In detail:

    Making a shared logger calls code within the multiprocessing module.
    This in turn tries to import the hmac module inside deliver_challenge().
    This hangs if it happens after a module fork.

    The only way around this is to only make calls to multiprocessing
    (i.e. make_shared_logger_and_proxy(...)) after the import phase of
    module loading.

    This python bug will be triggered if your make_shared_logger_and_proxy()
    call is at global scope in a module (i.e. not __main__) and only for
    python version 3.2

    888888888888888888888888888888888888888888888888888888888888888888888888888

""")
    sys.exit()

(logger_proxy,
 logging_mutex) = make_shared_logger_and_proxy(setup_std_shared_logger,
                                               "my_logger", args)


#
#    task1
#
@originate(input_file_names, logger_proxy, logging_mutex)
def task1(outfile, logger_proxy, logging_mutex):
    write_input_output_filenames_to_output(
        None, outfile, logger_proxy, logging_mutex)


#
#    task2
#
@transform(task1, suffix(".1"), ".2", logger_proxy, logging_mutex)
def task2(infile, outfile, logger_proxy, logging_mutex):
    write_input_output_filenames_to_output(
        infile, outfile, logger_proxy, logging_mutex)


#
#    task3
#
@transform(task2, suffix(".2"), ".3", logger_proxy, logging_mutex)
def task3(infile, outfile, logger_proxy, logging_mutex):
    """
    Third task
    """
    write_input_output_filenames_to_output(
        infile, outfile, logger_proxy, logging_mutex)


#
#    task4
#
@merge(task3, final_file_name, logger_proxy, logging_mutex)
def task4(infile, outfile, logger_proxy, logging_mutex):
    """
    Fourth task
    """
    write_input_output_filenames_to_output(
        infile, outfile, logger_proxy, logging_mutex)


class Test_ruffus(unittest.TestCase):
    def setUp(self):
        self.tearDown()
        try:
            os.makedirs(tempdir)
            #sys.stderr.write("    Created %s\n" % tempdir)
        except:
            pass

    def tearDown(self):
        try:
            shutil.rmtree(tempdir)
            #sys.stderr.write("    Removed %s\n" % tempdir)
            pass
        except:
            pass

    def test_simpler(self):
        pipeline_run(multiprocess=500, verbose=0, pipeline="main")

    def test_newstyle_simpler(self):
        test_pipeline = Pipeline("test")
        test_pipeline.originate(task1, input_file_names, extras=[
                                logger_proxy, logging_mutex])
        test_pipeline.transform(task2, task1, suffix(
            ".1"), ".2", extras=[logger_proxy, logging_mutex])
        test_pipeline.transform(task3, task2, suffix(
            ".2"), ".3", extras=[logger_proxy, logging_mutex])
        test_pipeline.merge(task4, task3, final_file_name,
                            extras=[logger_proxy, logging_mutex])
        #test_pipeline.merge(task4, task3, final_file_name, extras = {"logger_proxy": logger_proxy, "logging_mutex": logging_mutex})
        test_pipeline.run(multiprocess=500, verbose=0)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
import logging.handlers
import logging
import re
"""
    Parse Ruffus pipelines using old fashioned decorator syntax
"""

################################################################################
#
#   parse_old_style_ruffus.py
#
#
#   Copyright (c) 25 February 2015 Leo Goodstadt
#   This is free software, and you are welcome to redistribute it
#   under the "MIT" license; See http://opensource.org/licenses/MIT for details.
#
#################################################################################

import sys
import os

# Use import path from ~/python_modules
if __name__ == '__main__':
    sys.path.insert(0, os.path.expanduser("~lg/src/oss/ruffus"))
    sys.path.append(os.path.expanduser("~lg/src/python_modules"))


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#   options


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888


if __name__ == '__main__':
    import argparse

    parser = argparse.ArgumentParser(description=__doc__)

    common_group = parser.add_argument_group('Common arguments')
    common_group.add_argument('--verbose', "-v", const=1, metavar="VERBOSITY", default=0, nargs='?', type=int,
                              help="Print more verbose messages for each additional verbose level.")
    common_group.add_argument(
        '--version', action='version', version='%(prog)s 1.0')
    common_group.add_argument("-L", "--log_file", metavar="FILE", type=str,
                              help="Name and path of log file")

    options = parser.parse_args()

    if not options.log_file:
        options.log_file = os.path.join("parse_old_style_ruffus.log")


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#   imports


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#from json import dumps
#from collections import defaultdict


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#   Functions


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#   Logger


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888


MESSAGE = 15
logging.addLevelName(MESSAGE, "MESSAGE")


def setup_std_logging(module_name, log_file, verbose):
    """
    set up logging using programme options
    """

    logger = logging.getLogger(module_name)

    # We are interesting in all messages
    logger.setLevel(logging.DEBUG)
    has_handler = False

    #
    #   log to file if that is specified
    #
    if log_file:
        handler = logging.FileHandler(log_file, delay=False)

        class stripped_down_formatter(logging.Formatter):
            def format(self, record):
                prefix = ""
                if not hasattr(self, "first_used"):
                    self.first_used = True
                    prefix = "\n" + self.formatTime(record, "%Y-%m-%d")
                    prefix += " %(name)s\n" % record.__dict__
                if record.levelname in ("INFO", "MESSAGE", "DEBUG"):
                    self._fmt = " %(asctime)s - %(message)s"
                else:
                    self._fmt = " %(asctime)s - %(levelname)-7s - %(message)s"
                return prefix + logging.Formatter.format(self, record)
        handler.setFormatter(stripped_down_formatter(
            "%(asctime)s - %(name)s - %(levelname)6s - %(message)s", "%H:%M:%S"))
        handler.setLevel(MESSAGE)
        logger.addHandler(handler)
        has_handler = True

    #
    #   log to stderr if verbose
    #
    if verbose:
        stderrhandler = logging.StreamHandler(sys.stderr)
        stderrhandler.setFormatter(logging.Formatter("    %(message)s"))
        stderrhandler.setLevel(logging.DEBUG)
        if log_file:
            class debug_filter(logging.Filter):
                """
                Ignore INFO messages
                """

                def filter(self, record):
                    return logging.INFO != record.levelno
            stderrhandler.addFilter(debug_filter())
        logger.addHandler(stderrhandler)
        has_handler = True

    #
    #   no logging
    #
    if not has_handler:
        class NullHandler(logging.Handler):
            """
            for when there is no logging
            """

            def emit(self, record):
                pass
        logger.addHandler(NullHandler())

    return logger


if __name__ == '__main__':

    #
    #   set up log: name = script name sans extension
    #
    module_name = os.path.splitext(os.path.basename(sys.argv[0]))[0]
    logger = setup_std_logging(module_name, options.log_file, options.verbose)

    #
    #   log programme parameters
    #
    logger.info(" ".join(sys.argv))


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#   Main logic


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

func_re = re.compile("^\s*def.*\(")


def get_decorators(line_num, decorated_lines, all_lines):
    for line in sys.stdin:
        line_num += 1
        mm = func_re.search(line)
        if mm:
            decorated_lines.append(line + "\n")
            all_lines.append(line)
            return line_num
        else:
            decorated_lines.append(line)

    raise Exception("Unterminated decorators %s" % (decorated_lines,))


decorator_re = re.compile("^\s*@")
no_white_space_re = re.compile("^[^#\s]")
if __name__ == '__main__':
    line_num = 1
    last_decorated_line_num = 1
    pipeline_insertion_line_num = 1
    decorated_lines = []
    all_lines = []
    for line in sys.stdin:
        line_num += 1
        # decorator start
        if decorator_re.search(line):
            decorated_lines.append(line)
            line_num = get_decorators(line_num, decorated_lines, all_lines)
            last_decorated_line_num = len(all_lines)
            continue
        if no_white_space_re.search(line):
            if pipeline_insertion_line_num < last_decorated_line_num:
                pipeline_insertion_line_num = len(all_lines)
        all_lines.append(line)

    for line in all_lines[0: pipeline_insertion_line_num]:
        sys.stdout.write(line)
    sys.stdout.write("\n" * 3)

    sys.stdout.write('pipeline = Pipeline.pipelines["main"]\n\n')

    for line in decorated_lines:
        sys.stdout.write(line)
    sys.stdout.write("\n" * 3)

    for line in all_lines[pipeline_insertion_line_num:]:
        sys.stdout.write(line)


#        for aa in active_if.py \
#        branching_dependencies.py \
#        cmdline.py \
#        collate.py \
#        empty_files_decorator.py \
#        exceptions.py \
#        file_name_parameters.py \
#        files_decorator.py \
#        filesre_combine.py \
#        filesre_split_and_combine.py \
#        follows_mkdir.py \
#        graphviz.py \
#        inputs_with_multiple_args_raising_exception.py \
#        job_completion_checksums.py \
#        job_history_with_exceptions.py \
#        mkdir.py \
#        N_x_M_and_collate.py \
#        pausing.py \
#        pipeline_printout_graph.py \
#        posttask_merge.py \
#        regex_error_messages.py \
#        ruffus_utility_parse_task_arguments.py \
#        ruffus_utility.py \
#        runtime_data.py \
#        softlink_uptodate.py \
#        split_and_combine.py \
#        split_regex_and_collate.py \
#        split_subdivide_checkpointing.py \
#        task_file_dependencies.py \
#        task_misc.py \
#        transform_add_inputs.py \
#        transform_inputs.py \
#        transform_with_no_re_matches.py \
#        tutorial7.py \
#        unicode_filenames.py \
#        verbosity.py; \
#        do echo test_$aa; ../parse_old_style_ruffus.py < test_$aa >| test_newstyle_$aa; done



#!/usr/bin/env python
from time import sleep
import sys

try:
    for i in range(50):
        sleep(1)
        sys.stderr.write("{}\n".format(i))
    sleep(1)
    sys.stderr.write("Done\n")
except:
    sys.stderr.write("Exception!!!!!!!!!!!!!!!!!!!!!\n")
    pass



#!/usr/bin/env python
from __future__ import print_function
import unittest
from ruffus import follows, pipeline_run, Pipeline, mkdir
import sys

"""
    test_follows_mkdir.py
"""


import os
tempdir = os.path.relpath(os.path.abspath(os.path.splitext(__file__)[0])) + "/"

# add grandparent to search path for testing
grandparent_dir = os.path.abspath(
    os.path.join(os.path.dirname(__file__), "..", ".."))
sys.path.insert(0, grandparent_dir)

# module name = script name without extension
module_name = os.path.splitext(os.path.basename(__file__))[0]


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#   Tasks


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888
directories = [os.path.abspath(tempdir + 'a'), tempdir + 'b']


@follows(mkdir(tempdir), mkdir(directories), mkdir(tempdir + 'c'), mkdir(tempdir + 'd', tempdir + 'e'), mkdir(tempdir + 'e'))
def task_which_makes_directories():
    pass


class Test_task_mkdir(unittest.TestCase):

    def setUp(self):
        """
        """
        pass

    def tearDown(self):
        """
        delete directories
        """
        for d in 'abcde':
            fullpath = os.path.join(os.path.dirname(__file__), tempdir, d)
            os.rmdir(fullpath)
        os.rmdir(tempdir)

    def test_mkdir(self):
        pipeline_run(multiprocess=10, verbose=0, pipeline="main")

        for d in 'abcde':
            fullpath = os.path.join(os.path.dirname(__file__), tempdir, d)
            self.assertTrue(os.path.exists(fullpath))

    def test_newstyle_mkdir(self):
        test_pipeline = Pipeline("test")
        test_pipeline.follows(task_which_makes_directories, mkdir(directories), mkdir(
            tempdir + 'c'), mkdir(tempdir + 'd', tempdir + 'e'), mkdir(tempdir + 'e'))
        test_pipeline.run(multiprocess=10, verbose=0)

        for d in 'abcde':
            fullpath = os.path.join(os.path.dirname(__file__), tempdir, d)
            self.assertTrue(os.path.exists(fullpath))


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
from __future__ import print_function
import shutil
import unittest
from ruffus import posttask, parallel, check_if_uptodate, follows, touch_file, pipeline_printout, pipeline_run
import sys

"""
Test check_if_uptodate

    Bug: @parallel sets @check_if_uptodate to None.
    Submitted by Jafar Taghiyar (jafar<dot>taghiyar<at>gmail.com)
    https://github.com/bunbun/ruffus/issues/53

"""
import os
tempdir = os.path.abspath(os.path.abspath(os.path.splitext(__file__)[0]))
exe_path = os.path.abspath(os.path.join(os.path.dirname(__file__), ".."))

# add grandparent to search path for testing
grandparent_dir = os.path.abspath(
    os.path.join(os.path.dirname(__file__), "..", ".."))
sys.path.insert(0, grandparent_dir)

# module name = script name without extension
module_name = os.path.splitext(os.path.basename(__file__))[0]


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#   imports


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

try:
    from StringIO import StringIO
except:
    from io import StringIO


def sentinel_file_exists(output_file):
    if not os.path.exists(output_file):
        return True, "Missing file %s" % output_file
    else:
        return False, "File %s exists" % output_file


@posttask(touch_file(os.path.join(tempdir, "task1_completed.flag")))
@parallel([[os.path.join(tempdir, "task1_completed.flag")]])
@check_if_uptodate(sentinel_file_exists)
def task1(x):
    pass


@follows(task1)
@posttask(touch_file(os.path.join(tempdir, "task2_completed.flag")))
@parallel([[os.path.join(tempdir, "task2_completed.flag")]])
@check_if_uptodate(sentinel_file_exists)
def task2(x):
    pass


class Test_ruffus(unittest.TestCase):

    def do_assertNotRegexpMatches(self, test_str, re_str):
        import re
        if re.search(re_str, test_str) is not None:
            raise AssertionError(
                "Regexp matched: %r found in %r" % (re_str, test_str))

    def do_assertRegexpMatches(self, test_str, re_str):
        import re
        if re.search(re_str, test_str) is None:
            # AssertionError:
            raise AssertionError(
                "Regexp didn't match: %r not found in %r" % (re_str, test_str))

    def setUp(self):
        try:
            shutil.rmtree(tempdir)
        except:
            pass
        os.makedirs(tempdir)

    def tearDown(self):
        try:
            shutil.rmtree(tempdir)
        except:
            pass

    def test_ruffus(self):
        # run first
        pipeline_run(verbose=0)
        # should now be out of date
        s = StringIO()
        pipeline_printout(s, verbose=5)

        ret = s.getvalue()
        try:
            self.do_assertRegexpMatches(
                ret, r"Tasks which are up-to-date:(\n\s*)*Task = 'test_check_if_uptodate.task1'(\n\s*)*Task = 'test_check_if_uptodate.task2'")
        except:
            print("\n\tOops: Both tasks should be up to date!!\n\n")
            raise
        try:
            self.do_assertNotRegexpMatches(
                ret, r"Jobs needs update:\s*No function to check if up-to-date")
        except:
            print("\n\tOops: @check_if_uptodate is not being picked up!!\n\n")
            raise


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
from __future__ import print_function
import unittest
import shutil
from ruffus.combinatorics import *
from ruffus.ruffus_utility import RUFFUS_HISTORY_FILE, CHECKSUM_FILE_TIMESTAMPS
from ruffus.ruffus_exceptions import RethrownJobError
from ruffus import pipeline_run, pipeline_printout, Pipeline, suffix, regex, formatter, originate, follows, merge, mkdir, posttask, subdivide, transform, collate, split
import ruffus
import re
import sys
import os

"""

    test_split_regex_and_collate.py

"""
JOBS_PER_TASK = 5

tempdir = os.path.relpath(os.path.abspath(os.path.splitext(__file__)[0])) + "/"

# add grandparent to search path for testing
grandparent_dir = os.path.abspath(
    os.path.join(os.path.dirname(__file__), "..", ".."))
sys.path.insert(0, grandparent_dir)

# module name = script name without extension
module_name = os.path.splitext(os.path.basename(__file__))[0]


# funky code to import by file name
parent_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), ".."))


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#   options


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

try:
    from StringIO import StringIO
except:
    from io import StringIO


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#   imports


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

# use simplejson in place of json for python < 2.6
try:
    import json
except ImportError:
    import simplejson
    json = simplejson

# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#   Main logic


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#
#   Three starting files
#
original_files = [tempdir + "/original_%d.fa" % d for d in range(3)]


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#   Tasks


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888
@mkdir(tempdir)
@originate(original_files)
def generate_initial_files(out_name):
    with open(out_name, 'w') as outfile:
        pass


#
#    split_fasta_file
#

@posttask(lambda: sys.stderr.write("\tSplit into %d files each\n" % JOBS_PER_TASK))
@subdivide(generate_initial_files,
           # match original files
           regex(r".*\/original_(\d+).fa"),
           [tempdir + r"/files.split.\1.success",                         # flag file for each original file
            tempdir + r"/files.split.\1.*.fa"],                           # glob pattern
           r"\1")                                                        # index of original file
def split_fasta_file(input_file, outputs, original_index):

    #
    # remove previous fasta files
    #
    success_flag = outputs[0]
    output_file_names = outputs[1:]
    for f in output_file_names:
        os.unlink(f)

    #
    # create as many files as we are simulating in JOBS_PER_TASK
    #
    for i in range(JOBS_PER_TASK):
        with open(tempdir + "/files.split.%s.%03d.fa" % (original_index, i), "w") as oo:
            pass

    with open(success_flag,  "w") as oo:
        pass


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888
#
#    align_sequences
#
@posttask(lambda: sys.stderr.write("\tSequences aligned\n"))
# fa -> aln
@transform(split_fasta_file, suffix(".fa"), ".aln")
def align_sequences(input_file, output_filename):
    with open(output_filename, "w") as oo:
        oo.write("%s\n" % output_filename)


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888
#
#    percentage_identity
#
@posttask(lambda: sys.stderr.write("\t%Identity calculated\n"))
@transform(align_sequences,             # find all results from align_sequences
           suffix(".aln"),             # replace suffix with:
           [r".pcid",  # .pcid suffix for the result
            r".pcid_success"])  # .pcid_success to indicate job completed
def percentage_identity(input_file, output_files):
    (output_filename, success_flag_filename) = output_files
    with open(output_filename, "w") as oo:
        oo.write("%s\n" % output_filename)
    with open(success_flag_filename, "w") as oo:
        pass


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888
#
#    combine_results
#
@posttask(lambda: sys.stderr.write("\tResults recombined\n"))
@collate(percentage_identity, regex(r".*files.split\.(\d+)\.\d+.pcid"),
         [tempdir + r"/\1.all.combine_results",
          tempdir + r"/\1.all.combine_results_success"])
def combine_results(input_files, output_files):
    """
    Combine all
    """
    (output_filename, success_flag_filename) = output_files
    with open(output_filename, "w") as out:
        for inp, flag in input_files:
            with open(inp) as ii:
                out.write(ii.read())
    with open(success_flag_filename, "w") as oo:
        pass


class Test_ruffus(unittest.TestCase):
    def setUp(self):
        import os
        try:
            shutil.rmtree(tempdir)
        except:
            pass
        os.makedirs(tempdir)
        for f in original_files:
            with open(f, "w") as p:
                pass

    def cleanup_tmpdir(self):
        os.system('rm -f %s %s' %
                  (os.path.join(tempdir, '*'), RUFFUS_HISTORY_FILE))

    # ___________________________________________________________________________
    #
    #   test product() pipeline_printout and pipeline_run
    # ___________________________________________________________________________
    def test_collate(self):
        self.cleanup_tmpdir()

        s = StringIO()
        pipeline_printout(s, [combine_results], verbose=5,
                          wrap_width=10000, pipeline="main")
        self.assertTrue(re.search(
            'Job needs update:.*Missing files.*', s.getvalue(), re.DOTALL) is not None)
        # print s.getvalue()

        pipeline_run([combine_results], verbose=0, pipeline="main")

    def test_newstyle_collate(self):
        """
        As above but create pipeline on the fly using object orientated syntax rather than decorators
        """

        #
        # Create pipeline on the fly, joining up tasks
        #
        test_pipeline = Pipeline("test")

        test_pipeline.originate(task_func=generate_initial_files,
                                output=original_files)\
            .mkdir(tempdir, tempdir+"/test")

        test_pipeline.subdivide(task_func=split_fasta_file,
                                input=generate_initial_files,
                                # match original files
                                filter=regex(r".*\/original_(\d+).fa"),
                                output=[tempdir + r"/files.split.\1.success",  # flag file for each original file
                                        tempdir + r"/files.split.\1.*.fa"],   # glob pattern
                                extras=[r"\1"])\
            .posttask(lambda: sys.stderr.write("\tSplit into %d files each\n" % JOBS_PER_TASK))

        test_pipeline.transform(task_func=align_sequences,
                                input=split_fasta_file,
                                filter=suffix(".fa"),
                                output=".aln")  \
            .posttask(lambda: sys.stderr.write("\tSequences aligned\n"))

        test_pipeline.transform(task_func=percentage_identity,
                                input=align_sequences,             # find all results from align_sequences
                                # replace suffix with:
                                filter=suffix(".aln"),
                                output=[r".pcid",  # .pcid suffix for the result
                                        r".pcid_success"]  # .pcid_success to indicate job completed
                                )\
            .posttask(lambda: sys.stderr.write("\t%Identity calculated\n"))

        test_pipeline.collate(task_func=combine_results,
                              input=percentage_identity,
                              filter=regex(r".*files.split\.(\d+)\.\d+.pcid"),
                              output=[tempdir + r"/\1.all.combine_results",
                                      tempdir + r"/\1.all.combine_results_success"])\
            .posttask(lambda: sys.stderr.write("\tResults recombined\n"))

        #
        # Cleanup, printout and run
        #
        self.cleanup_tmpdir()
        s = StringIO()
        test_pipeline.printout(s, [combine_results],
                               verbose=5, wrap_width=10000)
        self.assertTrue(re.search(
            'Job needs update:.*Missing files.*', s.getvalue(), re.DOTALL) is not None)
        test_pipeline.run(verbose=0)

    # ___________________________________________________________________________
    #
    #   cleanup
    # ___________________________________________________________________________
    def tearDown(self):
        shutil.rmtree(tempdir)


#
#   Necessary to protect the "entry point" of the program under windows.
#       see: http://docs.python.org/library/multiprocessing.html#multiprocessing-programming
#
if __name__ == '__main__':
    unittest.main()



from ruffus import *
from ruffus.drmaa_wrapper import *
import drmaa
drmaa_session = drmaa.Session()
drmaa_session.initialize()
logger, logger_mutex = cmdline.setup_logging("me", "test.log", 1)
job_queue_name = "short.qb"
job_other_options = '-P mott-flint.prjb'
run_job_using_drmaa("ls", job_queue_name, None, "test_job_name",
                    job_other_options, logger, drmaa_session)



#!/usr/bin/env python
from __future__ import print_function
import unittest
from ruffus.ruffus_utility import *
from ruffus import *

################################################################################
#
#   test_ruffus_utility.py
#
#################################################################################
"""
    test_ruffus_utility.py
"""


import os
import sys

# add grandparent to search path for testing
grandparent_dir = os.path.abspath(
    os.path.join(os.path.dirname(__file__), "..", ".."))
sys.path.insert(0, grandparent_dir)

# module name = script name without extension
module_name = os.path.splitext(os.path.basename(__file__))[0]


# _________________________________________________________________________________________

#   get_nested_tasks_or_globs

# _________________________________________________________________________________________
class Test_get_nested_tasks_or_globs(unittest.TestCase):
    def setUp(self):
        exe_path = os.path.split(os.path.abspath(sys.argv[0]))[0]
        os.chdir(exe_path)

    #       self.assertEqual(self.seq, range(10))
    #       self.assertTrue(element in self.seq)
    #       self.assertRaises(ValueError, random.sample, self.seq, 20)

    def check_equal(self, a, b):
        self.assertEqual(get_nested_tasks_or_globs(a), b)

    def test_get_nested_tasks_or_globs(self):

        #
        # test strings
        #
        self.check_equal("test", (list(), set(), set()))
        self.check_equal([("test1",), "test2", 3], (list(), set(), set()))

        #
        # test missing
        #
        self.check_equal((1, 3, [5]), (list(), set(), set()))
        self.check_equal(None, (list(), set(), set()))

        #
        # test glob
        #
        self.check_equal([("test1.*",), "test?2", 3],
                         (list(), set(['test1.*', 'test?2']), set()))

        #
        # test glob and string
        #
        self.check_equal([("test*1",), (("test3",),), "test2",
                          3], (list(), set(['test*1']), set()))

        #
        # test task function
        #
        self.check_equal(is_glob, (list([is_glob]), set([]), set()))
        self.check_equal([is_glob, [1, "this", ["that*", 5]], [(get_strings_in_flattened_sequence,)]], (
            [is_glob, get_strings_in_flattened_sequence], set(["that*"]), set()))
        #
        # test wrapper
        #
        self.check_equal(output_from(is_glob, ["what", 7], 5), ([
                         is_glob, "what"], set([]), set()))

# _________________________________________________________________________________________

#   replace_placeholders_with_tasks_in_input_params

# _________________________________________________________________________________________


class Test_replace_placeholders_with_tasks_in_input_params(unittest.TestCase):
    def setUp(self):
        exe_path = os.path.split(os.path.abspath(sys.argv[0]))[0]
        os.chdir(exe_path)

    #       self.assertEqual(self.seq, range(10))
    #       self.assertTrue(element in self.seq)
    #       self.assertRaises(ValueError, random.sample, self.seq, 20)

    def check_equal(self, a, b, d):
        self.assertEqual(
            replace_placeholders_with_tasks_in_input_params(a, d), b)

    def test_replace_placeholders_with_tasks_in_input_params(self):
        func_or_name_to_task = {is_glob: "FF is_glob", "what": "FF what",
                                get_strings_in_flattened_sequence: "FF get_strings_in_flattened_sequence"}

        #
        # test strings
        #
        self.check_equal("test", "test", func_or_name_to_task)
        self.check_equal([("test1",), "test2", 3],
                         [("test1",), "test2", 3],
                         func_or_name_to_task)

        #
        # test missing
        #
        self.check_equal((1, 3, [5]), (1, 3, [5]), func_or_name_to_task)
        self.check_equal(None, None, func_or_name_to_task)

        #
        # test task function
        #
        self.check_equal(is_glob, "FF is_glob", func_or_name_to_task)
        self.check_equal([is_glob, [1, "this", ["that*", 5]], [(get_strings_in_flattened_sequence,)]],
                         ["FF is_glob", [1, "this", ["that*", 5]],
                             [("FF get_strings_in_flattened_sequence",)]],
                         func_or_name_to_task)
        #
        # test wrapper
        #
        self.check_equal(output_from(is_glob, ["what", 7], 5),
                         ["FF is_glob", ["FF what", 7], 5],
                         func_or_name_to_task)
        self.check_equal(output_from(is_glob),
                         "FF is_glob",
                         func_or_name_to_task)

        self.check_equal([1, output_from(is_glob, ["what", 7], 5)],
                         [1, "FF is_glob", ["FF what", 7], 5],
                         func_or_name_to_task)

        self.check_equal([1, output_from(is_glob), ["what", 7], 5],
                         [1, "FF is_glob", ["what", 7], 5],
                         func_or_name_to_task)


# _________________________________________________________________________________________

#   non_str_sequence

# _________________________________________________________________________________________
class Test_non_str_sequence(unittest.TestCase):

    def test_non_str_sequence(self):
        """
            non_str_sequence()
        """
        test_str1 = "asfas"

        class inherited_str (str):
            #
            # use __new__ instead of init because str is immutable
            #
            def __new__(cls, a):
                obj = super(inherited_str, cls).__new__(inherited_str, a)
                return obj

        test_str2 = inherited_str("test")

        class inherited_list (list):
            def __init__(self, *param):
                list.__init__(self, *param)
        test_str3 = list(test_str1)
        test_str4 = inherited_list(test_str2)
        self.assertTrue(not non_str_sequence(test_str1))
        self.assertTrue(not non_str_sequence(test_str2))
        self.assertTrue(non_str_sequence(test_str3))
        self.assertTrue(non_str_sequence(test_str4))

# _________________________________________________________________________________________

#   get_strings_in_flattened_sequence

# _________________________________________________________________________________________


class Test_get_strings_in_flattened_sequence(unittest.TestCase):

    def test_get_strings_in_flattened_sequence(self):
        """
            get_strings_in_flattened_sequence()
        """
        class inherited_str (str):
            #
            # use __new__ instead of init because str is immutable
            #
            def __new__(cls, a):
                obj = super(inherited_str, cls).__new__(inherited_str, a)
                return obj

        class inherited_list (list):
            def __init__(self, *param):
                list.__init__(self, *param)

        self.assertEqual(get_strings_in_flattened_sequence("one"), ["one"])
        self.assertEqual(get_strings_in_flattened_sequence(
            ["one", "two"]), ["one", "two"])
        self.assertEqual(get_strings_in_flattened_sequence(
            ["one", 1, "two"]), ["one", "two"])
        self.assertEqual(get_strings_in_flattened_sequence(
            ["one", [1, ["two"]]]), ["one", "two"])
        self.assertEqual(get_strings_in_flattened_sequence(
            [inherited_str("one"), [1, ["two"]]]), [inherited_str("one"), "two"])
        self.assertEqual(get_strings_in_flattened_sequence(inherited_list([inherited_str("one"), [1, ["two"]]])),
                         inherited_list([inherited_str("one"), "two"]))


# _________________________________________________________________________________________

#   get_first_strings_in_nested_sequence

# _________________________________________________________________________________________
class Test_get_first_strings_in_nested_sequence(unittest.TestCase):

    def test_get_first_strings_in_nested_sequence(self):
        """
            get_first_strings_in_nested_sequence()
        """
        class inherited_str (str):
            #
            # use __new__ instead of init because str is immutable
            #
            def __new__(cls, a):
                obj = super(inherited_str, cls).__new__(inherited_str, a)
                return obj

        class inherited_list (list):
            def __init__(self, *param):
                list.__init__(self, *param)

        self.assertEqual(get_strings_in_flattened_sequence("one"), ["one"])
        self.assertEqual(get_strings_in_flattened_sequence(
            ["one", "two"]), ["one", "two"])
        self.assertEqual(get_strings_in_flattened_sequence(
            ["one", 1, "two"]), ["one", "two"])
        self.assertEqual(get_strings_in_flattened_sequence(
            ["one", [1, ["two"]]]), ["one", "two"])
        self.assertEqual(get_strings_in_flattened_sequence(
            [inherited_str("one"), [1, ["two"]]]), [inherited_str("one"), "two"])
        self.assertEqual(get_strings_in_flattened_sequence(inherited_list([inherited_str("one"), [1, ["two"]]])),
                         inherited_list([inherited_str("one"), "two"]))
        self.assertEqual(get_strings_in_flattened_sequence(
            ["one", [1, ["two"], "three"]]), ["one", "two", "three"])
        d = {"four": 4}
        self.assertEqual(get_strings_in_flattened_sequence(
            ["one", [1, [d, "two"], "three"]]), ["one", "two", "three"])
        self.assertEqual(get_strings_in_flattened_sequence(None), [])
        self.assertEqual(get_strings_in_flattened_sequence([]), [])
        self.assertEqual(get_strings_in_flattened_sequence([1, 2, 3, d]), [])


# _________________________________________________________________________________________

#   Test_compile_regex

# _________________________________________________________________________________________
class Test_compile_regex (unittest.TestCase):
    def test_compile_regex(self):
        compile_regex("Dummy Task", regex(".*"), Exception, "test1")

        # bad regex
        self.assertRaises(Exception, compile_regex,
                          "Dummy Task", regex(".*)"), Exception, "test1")
        try:
            compile_regex("Dummy Task", regex(".*)"), Exception, "test1")
        except Exception as e:
            self.assertTrue(e.args == ('Dummy Task', "test1: regular expression regex('.*)') is malformed\n[sre_constants.error: (unbalanced parenthesis at position 2)]") or
                            e.args == ('Dummy Task', "test1: regular expression regex('.*)') is malformed\n[sre_constants.error: (unbalanced parenthesis)]") or
                            e.args == ('Dummy Task', "test1: regular expression regex('.*)') is malformed\n[re.error: (unbalanced parenthesis at position 2)]"))

        # bad number of items regex
        self.assertRaises(Exception, compile_regex, "Dummy Task",
                          regex(".*", "o"), Exception, "test1")
        try:
            compile_regex("Dummy Task", regex(".*", "o"), Exception, "test1")
        except Exception as e:
            self.assertEqual(
                e.args, ('Dummy Task', "test1: regex('.*', 'o') is malformed\nregex(...) should only be used to wrap a single regular expression string"))

        # 0 number of items regex
        self.assertRaises(Exception, compile_regex,
                          "Dummy Task", regex(), Exception, "test1")
        try:
            compile_regex("Dummy Task", regex(), Exception, "test1")
        except Exception as e:
            self.assertEqual(
                e.args, ('Dummy Task', 'test1: regex() is malformed\nregex(...) should only be used to wrap a single regular expression string'))

        # bad number of items suffix
        self.assertRaises(Exception, compile_suffix, "Dummy Task",
                          suffix(".*", "o"), Exception, "test1")
        try:
            compile_suffix("Dummy Task", suffix(".*", "o"), Exception, "test1")
        except Exception as e:
            self.assertEqual(
                e.args, ('Dummy Task', "test1: suffix('('.*', 'o')') is malformed.\nsuffix(...) should only be used to wrap a single string matching the suffices of file names"))

        # 0 number of items suffix
        self.assertRaises(Exception, compile_suffix,
                          "Dummy Task", suffix(), Exception, "test1")
        try:
            compile_suffix("Dummy Task", suffix(), Exception, "test1")
        except Exception as e:
            self.assertEqual(
                e.args, ('Dummy Task', 'test1: suffix() is malformed.\nsuffix(...) should be used to wrap a string matching the suffices of file names'))

# _________________________________________________________________________________________

#   Test_check_files_io_parameters

# _________________________________________________________________________________________


class Test_check_files_io_parameters (unittest.TestCase):
    def test_check_files_io_parameters(self):

        class t_fake_task(object):
            def __init__(self):
                self._action_type = None
                self._name = "fake task"
        fake_task = t_fake_task()

        single_job_params = [["input", "output"]]
        multiple_job_params = [["input1", "output1"], ["input2", "output2"]]

        check_files_io_parameters(
            fake_task, single_job_params, error_task_files)
        check_files_io_parameters(
            fake_task, multiple_job_params, error_task_files)

        # Bad format
        bad_single_job_params = ["input", "output"]
        self.assertRaises(error_task_files, check_files_io_parameters,
                          fake_task, bad_single_job_params, error_task_files)

        # Missing output files for job
        bad_multiple_job_params = [["input1", "output1"], ["input2"]]
        self.assertRaises(error_task_files, check_files_io_parameters,
                          fake_task, bad_multiple_job_params, error_task_files)

        # Missing input files for job
        bad_multiple_job_params = [["input1", "output1"], []]
        self.assertRaises(error_task_files, check_files_io_parameters,
                          fake_task, bad_multiple_job_params, error_task_files)

        # Input or output file parameters should contain at least one or more file names strings
        #bad_multiple_job_params = [[1, 2]]
        #self.assertRaises(error_task_files, check_files_io_parameters, fake_task, bad_multiple_job_params, error_task_files)

# _________________________________________________________________________________________

#   Test_get_first_string_in_nested_sequence

# _________________________________________________________________________________________


class Test_get_first_string_in_nested_sequence (unittest.TestCase):
    def test_get_first_string_in_nested_sequence(self):

        self.assertEqual(get_first_string_in_nested_sequence("a"),  "a")
        self.assertEqual(get_first_string_in_nested_sequence(None),  None)
        self.assertEqual(get_first_string_in_nested_sequence(1),  None)
        self.assertEqual(get_first_string_in_nested_sequence((1, 2)),  None)
        self.assertEqual(
            get_first_string_in_nested_sequence((1, 2, "a")),  "a")
        self.assertEqual(
            get_first_string_in_nested_sequence((1, 2, "a")),  "a")
        self.assertEqual(get_first_string_in_nested_sequence(
            (1, [2, "b"], "a")),  "b")
        self.assertEqual(get_first_string_in_nested_sequence(
            (1, set([2, "b"]), "a")),  "b")

# _________________________________________________________________________________________

#   Test_check_parallel_parameters

# _________________________________________________________________________________________


class Test_check_parallel_parameters (unittest.TestCase):
    def test_check_parallel_parameters(self):

        class t_fake_task(object):
            def __init__(self):
                self._action_type = None
                self._name = "fake task"
        fake_task = t_fake_task()

        single_job_params = [["input", "output"]]
        multiple_job_params = [["input1", "output1"], ["input2", "output2"]]

        check_parallel_parameters(
            fake_task, single_job_params, error_task_files)
        check_parallel_parameters(
            fake_task, multiple_job_params, error_task_files)

        # Bad format
        bad_single_job_params = ["input", "output"]
        self.assertRaises(error_task_parallel, check_parallel_parameters,
                          fake_task, bad_single_job_params, error_task_parallel)

# _________________________________________________________________________________________

#   expand_nested_tasks_or_globs

# _________________________________________________________________________________________


class Test_expand_nested_tasks_or_globs(unittest.TestCase):
    def setUp(self):
        exe_path = os.path.split(os.path.abspath(sys.argv[0]))[0]
        os.chdir(exe_path)
        t1 = task.Task(print, "module.func1")
        t2 = task.Task(print, "module.func2")
        t3 = task.Task(print, "module.func3")
        self.tasks = [t1, t2, t3]

    #       self.assertEqual(self.seq, range(10))
    #       self.assertTrue(element in self.seq)
    #       self.assertRaises(ValueError, random.sample, self.seq, 20)

    def check_equal(self, a, b):

        tasks, globs, runtime_data_names = get_nested_tasks_or_globs(a)
        func_or_name_to_task = dict(list(
            zip((non_str_sequence, get_strings_in_flattened_sequence, "what"), self.tasks)))

        task_or_glob_to_files = dict()
        # for f in func_or_name_to_task:
        #    print f, task_or_glob_to_files[func_or_name_to_task[f]]

        task_or_glob_to_files[self.tasks[0]] = [
            "t1a", "t1b"]       # non_str_sequence
        # get_strings_in_flattened_sequence
        task_or_glob_to_files[self.tasks[1]] = ["t2"]
        task_or_glob_to_files[self.tasks[2]] = ["t3"]               # "what"
        task_or_glob_to_files["that*"] = ["that1", "that2"]
        task_or_glob_to_files["test*1"] = ["test11", "test21"]
        task_or_glob_to_files["test1.*"] = ["test1.1", "test1.2"]
        task_or_glob_to_files["test?2"] = ["test12"]

        param_a = replace_placeholders_with_tasks_in_input_params(
            a, func_or_name_to_task)
        self.assertEqual(expand_nested_tasks_or_globs(
            param_a, task_or_glob_to_files), b)

    def test_expand_nested_tasks_or_globs(self):

        #
        # test strings
        #
        self.check_equal("test", "test")
        self.check_equal([("test1",), "test2", 3], [("test1",), "test2", 3])

        #
        # test missing
        #
        self.check_equal(None, None)

        #
        # test glob
        #
        self.check_equal([("test1.*",), "test?2", 3],
                         [("test1.1", "test1.2"), "test12", 3])
        self.check_equal(["test1.*", "test?2", 3],
                         ["test1.1", "test1.2", "test12", 3])

        #
        # test glob and string
        #
        self.check_equal([("test*1",), (("test3",),), "test2", 3],
                         [("test11", "test21"), (("test3",),), "test2", 3])

        #
        # test task function
        #
        self.check_equal(non_str_sequence, ["t1a", "t1b"])
        self.check_equal(get_strings_in_flattened_sequence, ["t2"])
        self.check_equal([get_strings_in_flattened_sequence,
                          non_str_sequence], ["t2", "t1a", "t1b"])
        self.check_equal([non_str_sequence, [1, "this", ["that*", 5]], [(get_strings_in_flattened_sequence,)]],
                         ['t1a', 't1b', [1, 'this', ['that1', 'that2', 5]], [('t2',)]])
        #
        # test wrapper
        #
        self.check_equal(output_from(non_str_sequence, ["what", 7], 5),
                         ['t1a', 't1b', ['t3', 7], 5])
#
#
# _________________________________________________________________________________________

#   Test_regex_replace

# _________________________________________________________________________________________


class Test_regex_replace (unittest.TestCase):
    def helper(self, data, result):
        regex_str = "([a-z]+)\.([a-z]+)\.([a-z]+)\.([a-z]+)"
        try_result = regex_replace("aaa.bbb.ccc.aaa",
                                   regex_str,
                                   re.compile(regex_str),
                                   data)
        self.assertEqual(try_result,  result)

    def test_regex_replace(self):
        self.helper(r"\3.\2.\1", "ccc.bbb.aaa")
        self.helper(None, None)
        self.helper(1, 1)
        self.helper([r"\3.\2.\1", 1], ["ccc.bbb.aaa", 1])
        # note set is constructed with substituted results!
        self.helper([r"\3.\2.\1", 1, (set([r"\1\2", r"\4\2", "aaabbb"]), "whatever", {1: 2, 3: 4})],
                    ['ccc.bbb.aaa', 1, (set(['aaabbb']), 'whatever', {1: 2, 3: 4})])


# _________________________________________________________________________________________

#   Test_path_decomposition

# _________________________________________________________________________________________
class Test_path_decomposition (unittest.TestCase):
    def helper(self, test_path, expected_result):
        try_result = path_decomposition(test_path)
        self.assertEqual(try_result,  expected_result)

    def test_path_decomposition(self):
        # normal path
        self.helper("/a/b/c/d/filename.txt",
                    {'basename': 'filename',
                        'ext':      '.txt',
                        'subpath':     ['/a/b/c/d', '/a/b/c', '/a/b', '/a', '/'],
                        'subdir': ['d', 'c', 'b', 'a', '/'],
                        'path':    '/a/b/c/d'
                     })
        # double slash
        self.helper("//a/filename.txt",
                    {'basename': 'filename',
                        'ext':      '.txt',
                        'subpath':     ['//a', '//'],
                        'path':    '//a',
                        'subdir': ['a', '//']
                     })
        # test no path
        self.helper("filename.txt",
                    {'basename': 'filename',
                        'ext':      '.txt',
                        'subpath':     [],
                        'path':   '',
                        'subdir': []
                     })
        # root
        self.helper("/filename.txt",
                    {'basename': 'filename',
                        'ext':      '.txt',
                        'path':     '/',
                        'subpath':     ['/'],
                        'subdir':   ['/']
                     })
        # unrooted
        self.helper("a/b/filename.txt",
                    {'basename': 'filename',
                        'ext':      '.txt',
                        'subpath':     ['a/b', 'a'],
                        'path':     'a/b',
                        'subdir':   ['b', 'a']
                     })
        # glob
        self.helper("/a/b/*.txt",
                    {'basename': '*',
                        'ext':      '.txt',
                        'path':     '/a/b',
                        'subpath':     ['/a/b', '/a', '/'],
                        'subdir':   ['b', 'a', '/']
                     })
        # no basename
        # extention becomes basename
        self.helper("/a/b/.txt",
                    {'basename': '.txt',
                        'ext':      '',
                        'path':     '/a/b',
                        'subpath':     ['/a/b', '/a', '/'],
                        'subdir':   ['b', 'a', '/']
                     })
        # no ext
        self.helper("/a/b/filename",
                    {'basename': 'filename',
                        'ext':      '',
                        'path':     '/a/b',
                        'subpath':     ['/a/b', '/a', '/'],
                        'subdir':   ['b', 'a', '/']
                     })
        # empty ext
        self.helper("/a/b/filename.",
                    {'basename': 'filename',
                        'ext':      '.',
                        'path':     '/a/b',
                        'subpath':  ['/a/b', '/a', '/'],
                        'subdir':   ['b', 'a', '/']
                     })
        # only path
        self.helper("/a/b/",
                    {'basename': '',
                        'ext':      '',
                        'path':     '/a/b',
                        'subpath':     ['/a/b', '/a', '/'],
                        'subdir':   ['b', 'a', '/']
                     })


class Test_apply_func_to_sequence (unittest.TestCase):
    def helper(self, test_seq, func, tuple_of_conforming_types, expected_result):
        try_result = apply_func_to_sequence(
            test_seq, func, tuple_of_conforming_types)
        self.assertEqual(try_result, expected_result)

    def test_apply_func_to_sequence(self):

        self.helper([
            ["saf", "sdfasf", 1],
            2,
            set([2, "odd"]),
            {1: 2},
            [
                ["sadf", 3]
            ]
        ],
            len, (str,),
            [
            [3, 6, 1],
            2,
            set([2, 3]),
            {1: 2},
            [
                [4, 3]
            ]
        ])

# _________________________________________________________________________________________

#   Test_parameter_list_as_string

# _________________________________________________________________________________________


class Test_parameter_list_as_string (unittest.TestCase):

    def test_conversion(self):

        self.assertEqual(parameter_list_as_string([1, 2, 3]),
                         '1, 2, 3')
        self.assertEqual(parameter_list_as_string([1, "2", 3]),
                         "1, '2', 3")
        self.assertEqual(parameter_list_as_string([1, None, 3]),
                         '1, None, 3')
        self.assertEqual(parameter_list_as_string([1, [2, 3], 3]),
                         '1, [2, 3], 3')
        self.assertEqual(parameter_list_as_string([1, [], 3]),
                         '1, [], 3')
        self.assertEqual(parameter_list_as_string(None),
                         '')
        self.assertRaises(TypeError, parameter_list_as_string)

#
# _________________________________________________________________________________________

#   Test_regex_match_str

# _________________________________________________________________________________________


class Test_regex_match_str (unittest.TestCase):

    def test_matches(self):

        # first string named and unamed captures, second string no captures
        test_str_list = ["aaa.bbb.ccc", "ddd.eee.fff"]
        compiled_regexes = ["aaa.(b+).(?P<CCC>c+)", "ddd.eee.fff"]
        results = [{0: 'aaa.bbb.ccc', 1: 'bbb',
                    2: 'ccc', 'CCC': 'ccc'}, {0: 'ddd.eee.fff'}]
        for ss, rr, result in zip(test_str_list, compiled_regexes, results):
            self.assertEqual(regex_matches_as_dict(ss, rr), result)

        # first string named and unamed captures, second string unnamed captures
        compiled_regexes = ["aaa.(b+).(?P<CCC>c+)", ".+(f)"]
        results = [{0: 'aaa.bbb.ccc', 1: 'bbb', 2: 'ccc',
                    'CCC': 'ccc'}, {0: 'ddd.eee.fff', 1: 'f'}]
        for ss, rr, result in zip(test_str_list, compiled_regexes, results):
            self.assertEqual(regex_matches_as_dict(ss, rr), result)

        # first string named and unamed captures, second string no capture
        compiled_regexes = ["aaa.(b+).(?P<CCC>c+)", ".+"]
        results = [{0: 'aaa.bbb.ccc', 1: 'bbb',
                    2: 'ccc', 'CCC': 'ccc'}, {0: 'ddd.eee.fff'}]
        for ss, rr, result in zip(test_str_list, compiled_regexes, results):
            self.assertEqual(regex_matches_as_dict(ss, rr), result)

        # first string named and unamed captures, second string None
        compiled_regexes = ["aaa.(b+).(?P<CCC>c+)", None]
        results = [{0: 'aaa.bbb.ccc', 1: 'bbb', 2: 'ccc', 'CCC': 'ccc'}, None]
        for ss, rr, result in zip(test_str_list, compiled_regexes, results):
            self.assertEqual(regex_matches_as_dict(ss, rr), result)

        # Both None
        compiled_regexes = []
        results = [None, None]
        for ss, rr, result in zip(test_str_list, compiled_regexes, results):
            self.assertEqual(regex_matches_as_dict(ss, rr), result)

        # first string named and unamed captures, second string Failed
        compiled_regexes = ["aaa.(b+).(?P<CCC>c+)", "PP"]
        results = [{0: 'aaa.bbb.ccc', 1: 'bbb', 2: 'ccc', 'CCC': 'ccc'}, False]
        for ss, rr, result in zip(test_str_list, compiled_regexes, results):
            self.assertEqual(regex_matches_as_dict(ss, rr), result)

        # first string named and unamed captures, second parameter number not string
        self.assertRaises(Exception, regex_matches_as_dict,
                          test_str_list[0], 6)


# _________________________________________________________________________________________

#   Test_path_decomposition

# _________________________________________________________________________________________
class Test_get_all_paths_components (unittest.TestCase):
    def helper(self, test_paths, regex_str, expected_result):
        try_result = get_all_paths_components(test_paths, regex_str)
        self.assertEqual(try_result,  expected_result)

    def test_get_all_paths_components(self):
        # no regex
        self.helper(["/a/b/c/sample1.bam"], None,
                    [
                        {'basename': 'sample1',
                         'ext': '.bam',
                         'subpath': ['/a/b/c', '/a/b', '/a', '/'],
                         'path': '/a/b/c',
                         'subdir': ['c', 'b', 'a', '/']
                         }
        ])

        # regex
        self.helper(["/a/b/c/sample1.bam"], [r"(.*)(?P<id>\d+)\..+"],
                    [
                        {
                            0: '/a/b/c/sample1.bam',
                            1: '/a/b/c/sample',
                            2: '1',
                            'id': '1',
                            'basename': 'sample1',
                            'ext': '.bam',
                            'subpath': ['/a/b/c', '/a/b', '/a', '/'],
                            'path': '/a/b/c',
                            'subdir': ['c', 'b', 'a', '/']
                        }
        ])
        # nameclash
        # "basename" overridden by named regular expression capture group
        self.helper(["/a/b/c/sample1.bam"], [r"(.*)(?P<basename>\d+)\..+"],
                    [
                        {
                            0: '/a/b/c/sample1.bam',
                            1: '/a/b/c/sample',
                            2: '1',
                            'basename': '1',
                            'ext': '.bam',
                            'subpath': ['/a/b/c', '/a/b', '/a', '/'],
                            'path': '/a/b/c',
                            'subdir': ['c', 'b', 'a', '/']
                        }
        ])

        # empty path
        self.helper([""], [r"(.*)(?P<basename>\d+)\..+"], [{}])
        self.helper(
            [""], [], [{'path': [], 'basename': '', 'ext': '', 'subdir': []}])
        # not matching regular expression
        self.helper(["/a/b/c/nonumber.txt"], [r"(.*)(?P<id>\d+)\..+"], [{}])
        # multiple paths
        self.helper(["/a/b/c/sample1.bam",
                     "dbsnp15.vcf",
                     "/test.txt"],
                    [
                        r"(.*)(?P<id>\d+)\..+",
                        r"(.*)(?P<id>\d+)\..+"],
                    [{
                        0:          '/a/b/c/sample1.bam',           # captured by index
                        1:          '/a/b/c/sample',                # captured by index
                        2:          '1',                            # captured by index
                        'id':       '1',                            # captured by name
                        'ext':      '.bam',
                        'subdir':   ['c', 'b', 'a', '/'],
                        'subpath':     ['/a/b/c', '/a/b', '/a', '/'],
                        'path': '/a/b/c',
                        'basename': 'sample1',
                    },
                        {
                            0: 'dbsnp15.vcf',                           # captured by index
                            1: 'dbsnp1',                                # captured by index
                            2: '5',                                     # captured by index
                            'id': '5',                                  # captured by name
                            'ext': '.vcf',
                            'subdir': [],
                            'subpath': [],
                            'path': '',
                            'basename': 'dbsnp15',
                    },

                        # no regular expression match
                        {
                            'ext': '.txt',
                            'subdir': ["/"],
                            'subpath': ["/"],
                            'path': '/',
                            'basename': 'test',
                    },
        ])
        # multiple paths : a single regular expression mismatch prevents any matches
        self.helper(["/a/b/c/sample1.bam",
                     "dbsnp15.vcf",
                     "/test.txt"],
                    [None,
                        r"(.*)(?P<id>\d+)\..+",
                        r"(.*)(?P<id>\d+)\..+"],
                    [{}, {}, {}])

        # _________________________________________________________________________________________

        #   Test_path_decomposition

        # _________________________________________________________________________________________

        class Test_get_all_paths_components (unittest.TestCase):
            def helper(self, test_paths, regex_str, expected_result):
                try_result = get_all_paths_components(test_paths, regex_str)
                self.assertEqual(try_result,  expected_result)

            def test_get_all_paths_components(self):
                # no regex
                self.helper(["/a/b/c/sample1.bam"], None,
                            [
                                {'basename': 'sample1',
                                 'ext': '.bam',
                                 'subpath': ['/a/b/c', '/a/b', '/a', '/'],
                                 'path': '/a/b/c',
                                 'subdir': ['c', 'b', 'a', '/']
                                 }
                ])

                # regex
                self.helper(["/a/b/c/sample1.bam"],
                            [r"(.*)(?P<id>\d+)\..+"],
                            [
                                {
                                    0: '/a/b/c/sample1.bam',
                                    1: '/a/b/c/sample',
                                    2: '1',
                                    'id': '1',
                                    'basename': 'sample1',
                                    'ext': '.bam',
                                    'subpath': ['/a/b/c', '/a/b', '/a', '/'],
                                    'path': '/a/b/c',
                                    'subdir': ['c', 'b', 'a', '/']
                                }
                ])
                # nameclash
                # "basename" overridden by named regular expression capture group
                self.helper(["/a/b/c/sample1.bam"], [r"(.*)(?P<basename>\d+)\..+"],
                            [
                                {
                                    0: '/a/b/c/sample1.bam',
                                    1: '/a/b/c/sample',
                                    'basename': '1',
                                    'ext': '.bam',
                                    'subpath': ['/a/b/c', '/a/b', '/a', '/'],
                                    'path': '/a/b/c',
                                    'subdir': ['c', 'b', 'a', '/']
                                }
                ])

                # empty path
                self.helper([""], [r"(.*)(?P<basename>\d+)\..+"], [{}])
                # not matching regular expression
                self.helper(["/a/b/c/nonumber.txt"],
                            [r"(.*)(?P<id>\d+)\..+"], [{}])
                # multiple paths
                self.helper(["/a/b/c/sample1.bam",
                             "dbsnp15.vcf",
                             "/test.txt"], [r"(.*)(?P<id>\d+)\..+"],
                            [{
                                0:          '/a/b/c/sample1.bam',           # captured by index
                                1:          '/a/b/c/sample',                # captured by index
                                2:          '1',                            # captured by index
                                'id':       '1',                            # captured by name
                                'ext':      '.bam',
                                'subdir':   ['c', 'b', 'a', '/'],
                                'subpath':     ['/a/b/c', '/a/b', '/a', '/'],
                                'path': '/a/b/c',
                                'basename': 'sample1',
                            },
                                {
                                    0: 'dbsnp15.vcf',                           # captured by index
                                    1: 'dbsnp1',                                # captured by index
                                    2: '5',                                     # captured by index
                                    'id': '5',                                  # captured by name
                                    'ext': '.vcf',
                                    'subdir': [],
                                    'subpath': [],
                                    'path': '',
                                    'basename': 'dbsnp15',
                            },

                                # no regular expression match
                                # everything fails!
                                {
                            }
                ])


#
# _________________________________________________________________________________________

#   Test_swap_nesting_order

# _________________________________________________________________________________________
class Test_swap_nesting_order (unittest.TestCase):

    def test_swap_nesting_order(self):
        orig_data = [
            {'a': 1, 'b': 2},
            {'a': 3, 'b': 4, 'c': 5}
        ]

        self.assertEqual(swap_nesting_order(orig_data),
                         ([],
                             {'a': {0: 1, 1: 3},
                              'c': {1: 5},
                              'b': {0: 2, 1: 4}})
                         )
        orig_data = [
            [{'a': 1, 'b': 2},
             {'a': 3, 'b': 4, 'c': 5}],
            [{'a': 6, 'b': 7},
             {'a': 8, 'b': 9, 'd': 10}]
        ]

        self.assertEqual(swap_doubly_nested_order(orig_data),
                         ([],
                          {'a':  {0: {0: 1, 1: 3}, 1: {0: 6, 1: 8}},
                           'c':  {0: {1: 5}},
                           'b':  {0: {0: 2, 1: 4}, 1: {0: 7, 1: 9}},
                           'd':  {1: {1: 10}}
                           }))


#
# _________________________________________________________________________________________

#   Test_shorten_filenames_encoder

# _________________________________________________________________________________________
class Test_shorten_filenames_encoder (unittest.TestCase):

    def setUp(self):
        import tempfile
        import os
        self.tempdir = tempfile.mkdtemp()
        subdir = self.tempdir + '/foo/bar/baz/bo'
        os.makedirs(subdir)
        os.chdir(subdir)

    def cleanUp(self):
        from shutil import rmtree
        rmtree(self.tempdir)

    def test_shorten_filenames_encoder(self):
        relative_path = os.path.abspath("../test1/something.py")
        absolute_path = "/a/long/path/to/oss/ruffus/ruffus/test/something.py"

        #
        # test relative path
        #
        self.assertEqual(shorten_filenames_encoder(relative_path, 4),
                         '../test1/something.py')

        # list of paths
        self.assertEqual(shorten_filenames_encoder([[relative_path, relative_path]] * 2 + [6], 4),
                         '[[../test1/something.py, ../test1/something.py], [../test1/something.py, ../test1/something.py], 6]')

        #
        # test full path
        #
        self.assertEqual(shorten_filenames_encoder(absolute_path, 4),
                         '.../ruffus/ruffus/test/something.py')

        # list of paths
        self.assertEqual(shorten_filenames_encoder([[absolute_path, absolute_path]] * 2 + [6], 4),
                         '[[.../ruffus/ruffus/test/something.py, .../ruffus/ruffus/test/something.py], '
                         '[.../ruffus/ruffus/test/something.py, .../ruffus/ruffus/test/something.py], 6]')


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
from __future__ import print_function
import ruffus.cmdline as cmdline
import shutil
from ruffus import add_inputs, suffix, mkdir, regex, Pipeline, output_from, touch_file
import ruffus
import sys

"""


        Demonstrates the new Ruffus syntax in version 2.6
"""

import os
tempdir = os.path.relpath(os.path.abspath(os.path.splitext(__file__)[0]))

# add grandparent to search path for testing
grandparent_dir = os.path.abspath(
    os.path.join(os.path.dirname(__file__), "..", ".."))
sys.path.insert(0, grandparent_dir)


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#   imports


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888


def touch(outfile):
    with open(outfile, "w"):
        pass


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#   Tasks


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888
def task_originate(o):
    """
    Makes new files
    """
    touch(o)


def task_m_to_1(i, o):
    """
    Merges files together
    """
    with open(o, "w") as o_file:
        for f in sorted(i):
            with open(f) as ii:
                o_file.write(f + "=" + ii.read() + "; ")


def task_1_to_1(i, o):
    """
    1 to 1 for transform
    """
    with open(o, "w") as o_file:
        with open(i) as ii:
            o_file.write(i + "+" + ii.read())


DEBUG_do_not_define_tail_task = False
DEBUG_do_not_define_head_task = False


#
#   Returns a fully formed sub pipeline useable as a building block
#

def make_pipeline1(pipeline_name,   # Pipelines need to have a unique name
                   starting_file_names):
    test_pipeline = Pipeline(pipeline_name)

    #   We can change the starting files later using
    #          set_input() for transform etc.
    #       or set_output() for originate
    #   But it can be more convenient to just pass this to the function making the pipeline
    #
    test_pipeline.originate(task_originate, starting_file_names)\
        .follows(mkdir(tempdir), mkdir(tempdir + "/testdir", tempdir + "/testdir2"))\
        .posttask(touch_file(tempdir + "/testdir/whatever.txt"))
    test_pipeline.transform(task_func=task_m_to_1,
                            name="add_input",
                            # Lookup Task from function name task_originate()
                            #   So long as this is unique in the pipeline
                            input=task_originate,
                            # requires an anchor from 3.7 onwards, see
                            # https://bugs.python.org/issue34982
                            filter=regex(r"^(.*)"),
                            add_inputs=add_inputs(
                                tempdir + "/testdir/whatever.txt"),
                            output=r"\1.22")
    test_pipeline.transform(task_func=task_1_to_1,
                            name="22_to_33",
                            # Lookup Task from Task name
                            #   Function name is not unique in the pipeline
                            input=output_from("add_input"),
                            filter=suffix(".22"),
                            output=".33")
    tail_task = test_pipeline.transform(task_func=task_1_to_1,
                                        name="33_to_44",
                                        # Ask Pipeline to lookup Task from Task name
                                        input=test_pipeline["22_to_33"],
                                        filter=suffix(".33"),
                                        output=".44")

    #   Set the tail task so that users of my sub pipeline can use it as a dependency
    #       without knowing the details of task names
    #
    #   Use Task() object directly without having to lookup
    test_pipeline.set_tail_tasks([tail_task])

    #   If we try to connect a Pipeline without tail tasks defined, we have to
    #       specify the exact task within the Pipeline.
    #   Otherwise Ruffus will not know which task we mean and throw an exception
    if DEBUG_do_not_define_tail_task:
        test_pipeline.set_tail_tasks([])

    # Set the head task so that users of my sub pipeline send input into it
    #   without knowing the details of task names
    test_pipeline.set_head_tasks([test_pipeline[task_originate]])

    return test_pipeline

#
#   Returns a fully formed sub pipeline useable as a building block
#


def make_pipeline2(pipeline_name="pipeline2"):
    test_pipeline2 = Pipeline(pipeline_name)
    test_pipeline2.transform(task_func=task_1_to_1,
                             # task name
                             name="44_to_55",
                             # placeholder: will be replaced later with set_input()
                             input=None,
                             filter=suffix(".44"),
                             output=".55")
    test_pipeline2.merge(task_func=task_m_to_1,
                         input=test_pipeline2["44_to_55"],
                         output=tempdir + "/final.output",)

    # Set head and tail
    test_pipeline2.set_tail_tasks([test_pipeline2[task_m_to_1]])
    if not DEBUG_do_not_define_head_task:
        test_pipeline2.set_head_tasks([test_pipeline2["44_to_55"]])

    return test_pipeline2


#   First two pipelines are created as separate instances by the make_pipeline1 function
pipeline1a = make_pipeline1(pipeline_name="pipeline1a", starting_file_names=[
                            tempdir + "/" + ss for ss in ("a.1", "b.1")])
pipeline1b = make_pipeline1(pipeline_name="pipeline1b", starting_file_names=[
                            tempdir + "/" + ss for ss in ("c.1", "d.1")])

#   The Third pipeline is a clone of pipeline1b
pipeline1c = pipeline1b.clone(new_name="pipeline1c")

#   Set the "originate" files for pipeline1c to ("e.1" and "f.1")
#       Otherwise they would use the original ("c.1", "d.1")
pipeline1c.set_output(output=[])
pipeline1c.set_output(output=[tempdir + "/" + ss for ss in ("e.1", "f.1")])

#   Join all pipeline1a-c to pipeline2
pipeline2 = make_pipeline2()
pipeline2.set_input(input=[pipeline1a, pipeline1b, pipeline1c])

parser = cmdline.get_argparse(
    description='Demonstrates the new Ruffus syntax in version 2.6')

parser.add_argument('--cleanup', "-C",
                    action="store_true",
                    help="Cleanup before and after.")

options = parser.parse_args(args=["--verbose=10"])
#  standard python logger which can be synchronised across concurrent Ruffus tasks
logger, logger_mutex = cmdline.setup_logging(
    __file__, options.log_file, options.verbose)

logger.debug("\tRuffus Version = " + ruffus.__version__)

if options.cleanup:
    try:
        shutil.rmtree(tempdir)
    except:
        pass

correct = False
# if we are not printing but running
if not options.just_print and \
   not options.flowchart and \
   not options.touch_files_only:

    #   Cleanup
    if options.cleanup:
        try:
            shutil.rmtree(tempdir)
        except:
            pass

    #   Run
    cmdline.run(options)

    # Check that the output reflecting the pipeline topology is correct.
    correct_output = '{tempdir}/a.1.55={tempdir}/a.1.44+{tempdir}/a.1.33+{tempdir}/a.1.22+{tempdir}/a.1=; {tempdir}/testdir/whatever.txt=; ; ' \
                     '{tempdir}/b.1.55={tempdir}/b.1.44+{tempdir}/b.1.33+{tempdir}/b.1.22+{tempdir}/b.1=; {tempdir}/testdir/whatever.txt=; ; ' \
                     '{tempdir}/c.1.55={tempdir}/c.1.44+{tempdir}/c.1.33+{tempdir}/c.1.22+{tempdir}/c.1=; {tempdir}/testdir/whatever.txt=; ; ' \
                     '{tempdir}/d.1.55={tempdir}/d.1.44+{tempdir}/d.1.33+{tempdir}/d.1.22+{tempdir}/d.1=; {tempdir}/testdir/whatever.txt=; ; ' \
                     '{tempdir}/e.1.55={tempdir}/e.1.44+{tempdir}/e.1.33+{tempdir}/e.1.22+{tempdir}/e.1=; {tempdir}/testdir/whatever.txt=; ; ' \
                     '{tempdir}/f.1.55={tempdir}/f.1.44+{tempdir}/f.1.33+{tempdir}/f.1.22+{tempdir}/f.1=; {tempdir}/testdir/whatever.txt=; ; '.format(
                         tempdir=tempdir)
    try:
        with open(tempdir + "/final.output") as real_output:
            real_output_str = real_output.read()
    except Exception as e:
        real_output_str = str(e) + "\n"
    if (correct_output != real_output_str):
        print("_" * 80 + "\n" +
              "      " * 25 + "Warning\n" +
              "_" * 80 + "\n" +
              "If you had run the whole of pipeline 2, "
              "you should have obtained:<\n\t%s>\n\n Rather than:<\n\t%s>\n\n"
              % (correct_output.replace("; ", ";\n\t"),
                  real_output_str.replace("; ", ";\n\t")) +
              "_" * 80, "\n",)
    else:
        logger.debug("\tAll Correct.\n")
        correct = True

    #
    #   Cleanup
    #
    if options.cleanup:
        try:
            shutil.rmtree(tempdir)
        except:
            pass


else:
    cmdline.run(options)
    correct = True


sys.exit(0 if correct else 1)



#!/usr/bin/env python
"""

    play_with_colours.py
    [--log_file PATH]
    [--verbose]

"""

################################################################################
#
#   test
#
#
#   Copyright (c) 7/13/2010 Leo Goodstadt
#
#   Permission is hereby granted, free of charge, to any person obtaining a copy
#   of this software and associated documentation files (the "Software"), to deal
#   in the Software without restriction, including without limitation the rights
#   to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#   copies of the Software, and to permit persons to whom the Software is
#   furnished to do so, subject to the following conditions:
#
#   The above copyright notice and this permission notice shall be included in
#   all copies or substantial portions of the Software.
#
#   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#   AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#   LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#   OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#   THE SOFTWARE.
#################################################################################

import sys, os


#88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#   options


#88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888


from optparse import OptionParser
import StringIO

parser = OptionParser(version="%play_with_colours 1.0",
                      usage = "\n\n    play_with_colours "
                                "--flowchart FILE [options] "
                                "[--colour_scheme_index INT ] "
                                "[--key_legend_in_graph]")

#
#   pipeline
#
parser.add_option("--flowchart", dest="flowchart",
                    metavar="FILE",
                    type="string",
                    help="Don't actually run any commands; just print the pipeline "
                         "as a flowchart.")
parser.add_option("--colour_scheme_index", dest="colour_scheme_index",
                    metavar="INTEGER",
                    type="int",
                    help="Index of colour scheme for flow chart.")
parser.add_option("--key_legend_in_graph", dest="key_legend_in_graph",
                    action="store_true", default=False,
                    help="Print out legend and key for dependency graph.")

(options, remaining_args) = parser.parse_args()
if not options.flowchart:
    raise Exception("Missing mandatory parameter: --flowchart.\n")




#88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#   imports


#88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

from ruffus import *
from ruffus.ruffus_exceptions import JobSignalledBreak


#88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#   Pipeline


#88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888


#
# up to date tasks
#
@check_if_uptodate (lambda : (False, ""))
def Up_to_date_task1(infile, outfile):
    pass

@check_if_uptodate (lambda : (False, ""))
@follows(Up_to_date_task1)
def Up_to_date_task2(infile, outfile):
    pass

@check_if_uptodate (lambda : (False, ""))
@follows(Up_to_date_task2)
def Up_to_date_task3(infile, outfile):
    pass


@check_if_uptodate (lambda : (False, ""))
@follows(Up_to_date_task3)
def Up_to_date_final_target(infile, outfile):
    pass


#
# Explicitly specified
#
@check_if_uptodate (lambda : (False, ""))
@follows(Up_to_date_task1)
def Explicitly_specified_task(infile, outfile):
    pass



#
# Tasks to run
#
@follows(Explicitly_specified_task)
def Task_to_run1(infile, outfile):
    pass


@follows(Task_to_run1)
def Task_to_run2(infile, outfile):
    pass

@follows(Task_to_run2)
def Task_to_run3(infile, outfile):
    pass

@check_if_uptodate (lambda : (False, ""))
@follows(Task_to_run2)
def Up_to_date_task_forced_to_rerun(infile, outfile):
    pass


#
# Final target
#
@follows(Up_to_date_task_forced_to_rerun, Task_to_run3)
def Final_target(infile, outfile):
    pass

#
# Ignored downstream
#
@follows(Final_target)
def Downstream_task1_ignored(infile, outfile):
    pass

@follows(Final_target)
def Downstream_task2_ignored(infile, outfile):
    pass











#88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#   Main logic


#88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888
from collections import defaultdict
custom_flow_chart_colour_scheme = defaultdict(dict)

#
#   Base chart on this overall colour scheme index
#
custom_flow_chart_colour_scheme["colour_scheme_index"] = options.colour_scheme_index

#
#   Overriding colours
#
if options.colour_scheme_index is None:
    custom_flow_chart_colour_scheme["Vicious cycle"]["linecolor"]                        = '"#FF3232"'
    custom_flow_chart_colour_scheme["Pipeline"]["fontcolor"]                             = '"#FF3232"'
    custom_flow_chart_colour_scheme["Key"]["fontcolor"]                                  = "black"
    custom_flow_chart_colour_scheme["Key"]["fillcolor"]                                  = '"#F6F4F4"'
    custom_flow_chart_colour_scheme["Task to run"]["linecolor"]                          = '"#0044A0"'
    custom_flow_chart_colour_scheme["Up-to-date"]["linecolor"]                           = "gray"
    custom_flow_chart_colour_scheme["Final target"]["fillcolor"]                         = '"#EFA03B"'
    custom_flow_chart_colour_scheme["Final target"]["fontcolor"]                         = "black"
    custom_flow_chart_colour_scheme["Final target"]["color"]                             = "black"
    custom_flow_chart_colour_scheme["Final target"]["dashed"]                            = 0
    custom_flow_chart_colour_scheme["Vicious cycle"]["fillcolor"]                        = '"#FF3232"'
    custom_flow_chart_colour_scheme["Vicious cycle"]["fontcolor"]                        = 'white'
    custom_flow_chart_colour_scheme["Vicious cycle"]["color"]                            = "white"
    custom_flow_chart_colour_scheme["Vicious cycle"]["dashed"]                           = 0
    custom_flow_chart_colour_scheme["Up-to-date task"]["fillcolor"]                      = '"#B8CC6E"'
    custom_flow_chart_colour_scheme["Up-to-date task"]["fontcolor"]                      = '"#006000"'
    custom_flow_chart_colour_scheme["Up-to-date task"]["color"]                          = '"#006000"'
    custom_flow_chart_colour_scheme["Up-to-date task"]["dashed"]                         = 0
    custom_flow_chart_colour_scheme["Down stream"]["fillcolor"]                          = "white"
    custom_flow_chart_colour_scheme["Down stream"]["fontcolor"]                          = "gray"
    custom_flow_chart_colour_scheme["Down stream"]["color"]                              = "gray"
    custom_flow_chart_colour_scheme["Down stream"]["dashed"]                             = 0
    custom_flow_chart_colour_scheme["Explicitly specified task"]["fillcolor"]            = "transparent"
    custom_flow_chart_colour_scheme["Explicitly specified task"]["fontcolor"]            = "black"
    custom_flow_chart_colour_scheme["Explicitly specified task"]["color"]                = "black"
    custom_flow_chart_colour_scheme["Explicitly specified task"]["dashed"]               = 0
    custom_flow_chart_colour_scheme["Task to run"]["fillcolor"]                          = '"#EBF3FF"'
    custom_flow_chart_colour_scheme["Task to run"]["fontcolor"]                          = '"#0044A0"'
    custom_flow_chart_colour_scheme["Task to run"]["color"]                              = '"#0044A0"'
    custom_flow_chart_colour_scheme["Task to run"]["dashed"]                             = 0
    custom_flow_chart_colour_scheme["Up-to-date task forced to rerun"]["fillcolor"]      = 'transparent'
    custom_flow_chart_colour_scheme["Up-to-date task forced to rerun"]["fontcolor"]      = '"#0044A0"'
    custom_flow_chart_colour_scheme["Up-to-date task forced to rerun"]["color"]          = '"#0044A0"'
    custom_flow_chart_colour_scheme["Up-to-date task forced to rerun"]["dashed"]         = 1
    custom_flow_chart_colour_scheme["Up-to-date Final target"]["fillcolor"]              = '"#EFA03B"'
    custom_flow_chart_colour_scheme["Up-to-date Final target"]["fontcolor"]              = '"#006000"'
    custom_flow_chart_colour_scheme["Up-to-date Final target"]["color"]                  = '"#006000"'
    custom_flow_chart_colour_scheme["Up-to-date Final target"]["dashed"]                 = 0

if __name__ == '__main__':
    pipeline_printout_graph (

                                    open(options.flowchart, "w"),
                                    # use flowchart file name extension to decide flowchart format
                                    #   e.g. svg, jpg etc.
                                    os.path.splitext(options.flowchart)[1][1:],

                                    # final targets
                                    [Final_target, Up_to_date_final_target],

                                    # Explicitly specified tasks
                                    [Explicitly_specified_task],

                                    # Do we want key legend
                                    no_key_legend   = not options.key_legend_in_graph,

                                    # Print all the task types whether used or not
                                    minimal_key_legend             = False,

                                    user_colour_scheme = custom_flow_chart_colour_scheme,
                                    pipeline_name = "Colour schemes")











#!/usr/bin/env python
from __future__ import print_function
import unittest
from ruffus import transform, Pipeline, pipeline_run, regex, inputs
import ruffus
import sys

"""

    test_inputs_with_multiple_args_raising_exception.py

        inputs with multiple arguments should raise an exception

"""


import os
tempdir = os.path.relpath(os.path.abspath(os.path.splitext(__file__)[0])) + "/"

# add grandparent to search path for testing
grandparent_dir = os.path.abspath(
    os.path.join(os.path.dirname(__file__), "..", ".."))
sys.path.insert(0, grandparent_dir)

# module name = script name without extension
module_name = os.path.splitext(os.path.basename(__file__))[0]


# funky code to import by file name
parent_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), ".."))


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#   imports


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#   Tasks


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888
try:
    @transform(None, regex(tempdir + "b"), inputs(tempdir + "a", tempdir + "b"), "task_1.output")
    def task_1(i, o):
        for f in o:
            open(f, 'w')
except ruffus.ruffus_exceptions.error_task_transform_inputs_multiple_args:
    print("\tExpected exception thrown 1")
except ruffus.ruffus_exceptions.error_inputs_multiple_args:
    print("\tExpected exception thrown 2")


def task_2(i, o):
    for f in o:
        open(f, 'w')


class Test_task_mkdir(unittest.TestCase):

    def setUp(self):
        """
        """
        pass

    def tearDown(self):
        """
        """
        pass

    def test_no_re_match(self):
        try:
            pipeline_run(multiprocess=10, verbose=0, pipeline="main")
        except:
            return
        raise Exception(
            "Inputs(...) with multiple arguments should have thrown an exception")

    def test_newstyle_no_re_match(self):
        try:
            test_pipeline = Pipeline("test")
            test_pipeline.transform(task_func=task_2,
                                    input=None,
                                    filter=regex(tempdir + "b"),
                                    replace_inputs=inputs(
                                        tempdir + "a", tempdir + "b"),
                                    output="task_1.output")
            test_pipeline.run(multiprocess=10, verbose=0)
        except ruffus.ruffus_exceptions.error_task_transform_inputs_multiple_args:
            print("\tExpected exception thrown 1")
            return
        except ruffus.ruffus_exceptions.error_inputs_multiple_args:
            print("\tExpected exception thrown 2")
            return
        raise Exception(
            "Inputs(...) with multiple arguments should have thrown an exception")


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
from __future__ import print_function
import re
import shutil
import unittest
from ruffus.ruffus_utility import RUFFUS_HISTORY_FILE
from ruffus import pipeline_run, pipeline_printout, Pipeline, formatter, output_from
import sys

"""

    test_combinatorics.py

        test product, combine, permute, combine_with_replacement

"""

import os
tempdir = os.path.relpath(os.path.abspath(os.path.splitext(__file__)[0]))
# sub-1s resolution in system?
one_second_per_job = None


# add grandparent to search path for testing
grandparent_dir = os.path.abspath(
    os.path.join(os.path.dirname(__file__), "..", ".."))
sys.path.insert(0, grandparent_dir)

# module name = script name without extension
module_name = os.path.splitext(os.path.basename(__file__))[0]


try:
    from StringIO import StringIO
except:
    from io import StringIO


def touch(outfile):
    with open(outfile, "w"):
        pass


# ___________________________________________________________________________
#
#   generate_initial_files1
# ___________________________________________________________________________
def generate_initial_files1(outfile):
    touch(outfile)

# ___________________________________________________________________________
#
#   generate_initial_files2
# ___________________________________________________________________________


def generate_initial_files2(outfile):
    touch(outfile)

# ___________________________________________________________________________
#
#   generate_initial_files3
# ___________________________________________________________________________


def generate_initial_files3(outfile):
    touch(outfile)

# ___________________________________________________________________________
#
#   check_product_task
# ___________________________________________________________________________


def check_product_task(infiles, outfile,
                       prefices,
                       subpath,
                       subdir):
    with open(outfile, "w") as p:
        p.write(prefices + ",")


# ___________________________________________________________________________
#
#   check_product_merged_task
# ___________________________________________________________________________
def check_product_merged_task(infiles, outfile):
    with open(outfile, "w") as p:
        for infile in sorted(infiles):
            with open(infile) as ii:
                p.write(ii.read())

# ___________________________________________________________________________
#
#   check_product_misspelt_capture_error_task
# ___________________________________________________________________________


def check_product_misspelt_capture_error_task(infiles, outfile):
    """
    FILE_PART mispelt as FILE_PART
    """
    touch(outfile)


# ___________________________________________________________________________
#
#   check_product_out_of_range_formatter_ref_error_task
# ___________________________________________________________________________
def check_product_out_of_range_formatter_ref_error_task(infiles, outfile, ignored_filter):
    """
    {path[2][0]} when len(path) == 1
    """
    touch(outfile)

# ___________________________________________________________________________
#
#   check_product_formatter_ref_index_error_task
# ___________________________________________________________________________


def check_product_formatter_ref_index_error_task(infiles, outfile, ignored_filter):
    """
    {path[0][0][1000} when len of the path string len(path[0][0]) < 1000
    """
    touch(outfile)

# ___________________________________________________________________________
#
#   check_combinations2_task
# ___________________________________________________________________________


def check_combinations2_task(infiles, outfile,
                             prefices,
                             subpath,
                             subdir):
    """
        Test combinations with k-tuple = 2
    """
    with open(outfile, "w") as outf:
        outf.write(prefices + ",")


def check_combinations2_merged_task(infiles, outfile):
    with open(outfile, "w") as p:
        for infile in sorted(infiles):
            with open(infile) as ii:
                p.write(ii.read())

# ___________________________________________________________________________
#
#   check_combinations3_task
# ___________________________________________________________________________


def check_combinations3_task(infiles, outfile,
                             prefices,
                             subpath,
                             subdir):
    """
        Test combinations with k-tuple = 3
    """
    with open(outfile, "w") as outf:
        outf.write(prefices + ",")


def check_combinations3_merged_task(infiles, outfile):
    with open(outfile, "w") as p:
        for infile in sorted(infiles):
            with open(infile) as ii:
                p.write(ii.read())


# ___________________________________________________________________________
#
#   check_permutations2_task
# ___________________________________________________________________________
def check_permutations2_task(infiles, outfile,
                             prefices,
                             subpath,
                             subdir):
    """
        Test permutations with k-tuple = 2
    """
    with open(outfile, "w") as outf:
        outf.write(prefices + ",")


def check_permutations2_merged_task(infiles, outfile):
    with open(outfile, "w") as p:
        for infile in sorted(infiles):
            with open(infile) as ii:
                p.write(ii.read())


# ___________________________________________________________________________
#
#   check_permutations3_task
# ___________________________________________________________________________
def check_permutations3_task(infiles, outfile,
                             prefices,
                             subpath,
                             subdir):
    """
        Test permutations with k-tuple = 3
    """
    with open(outfile, "w") as outf:
        outf.write(prefices + ",")


def check_permutations3_merged_task(infiles, outfile):
    with open(outfile, "w") as p:
        for infile in sorted(infiles):
            with open(infile) as ii:
                p.write(ii.read())


# ___________________________________________________________________________
#
#   check_combinations_with_replacement2_task
# ___________________________________________________________________________
def check_combinations_with_replacement2_task(infiles, outfile,
                                              prefices,
                                              subpath,
                                              subdir):
    """
        Test combinations_with_replacement with k-tuple = 2
    """
    with open(outfile, "w") as outf:
        outf.write(prefices + ",")


def check_combinations_with_replacement2_merged_task(infiles, outfile):
    with open(outfile, "w") as p:
        for infile in sorted(infiles):
            with open(infile) as ii:
                p.write(ii.read())


# ___________________________________________________________________________
#
#   check_combinations_with_replacement3_task
# ___________________________________________________________________________
def check_combinations_with_replacement3_task(infiles, outfile,
                                              prefices,
                                              subpath,
                                              subdir):
    """
        Test combinations_with_replacement with k-tuple = 3
    """
    with open(outfile, "w") as outf:
        outf.write(prefices + ",")


def check_combinations_with_replacement3_merged_task(infiles, outfile):
    with open(outfile, "w") as p:
        for infile in sorted(infiles):
            with open(infile) as ii:
                p.write(ii.read())


def cleanup_tmpdir():
    os.system('rm -f %s %s' %
              (os.path.join(tempdir, '*'), RUFFUS_HISTORY_FILE))


test_pipeline1 = Pipeline("test1")
test_pipeline2 = Pipeline("test2")
gen_task1 = test_pipeline1.originate(task_func=generate_initial_files1,
                                     name="WOWWWEEE",
                                     output=[tempdir + "/" + prefix + "_name.tmp1" for prefix in "abcd"])
test_pipeline1.originate(task_func=generate_initial_files2,
                         output=[tempdir + "/e_name.tmp1", tempdir + "/f_name.tmp1"])
test_pipeline1.originate(task_func=generate_initial_files3,
                         output=[tempdir + "/g_name.tmp1", tempdir + "/h_name.tmp1"])
test_pipeline1.product(task_func=check_product_task,
                       input=[tempdir + "/" + prefix +
                              "_name.tmp1" for prefix in "abcd"],
                       filter=formatter(".*/(?P<FILE_PART>.+).tmp1$"),
                       input2=generate_initial_files2,
                       filter2=formatter(),
                       input3=generate_initial_files3,
                       filter3=formatter(r"tmp1$"),
                       output="{path[0][0]}/{FILE_PART[0][0]}.{basename[1][0]}.{basename[2][0]}.tmp2",
                       extras=["{basename[0][0][0]}{basename[1][0][0]}{basename[2][0][0]}",       # extra: prefices only (abcd etc)
                               # extra: path for 2nd input, 1st file
                               "{subpath[0][0][0]}",
                               "{subdir[0][0][0]}"]).follows("WOWWWEEE").follows(gen_task1).follows(generate_initial_files1).follows("generate_initial_files1")
test_pipeline1.merge(task_func=check_product_merged_task,
                     input=check_product_task,
                     output=tempdir + "/merged.results")
test_pipeline1.product(task_func=check_product_misspelt_capture_error_task,
                       input=gen_task1,
                       filter=formatter(".*/(?P<FILE_PART>.+).tmp1$"),
                       output="{path[0][0]}/{FILEPART[0][0]}.tmp2")
test_pipeline1.product(task_func=check_product_out_of_range_formatter_ref_error_task,
                       input=generate_initial_files1,  #
                       filter=formatter(".*/(?P<FILE_PART>.+).tmp1$"),
                       output="{path[2][0]}/{basename[0][0]}.tmp2",
                       extras=["{FILE_PART[0][0]}"])
test_pipeline1.product(task_func=check_product_formatter_ref_index_error_task,
                       input=output_from("generate_initial_files1"),
                       filter=formatter(".*/(?P<FILE_PART>.+).tmp1$"),
                       output="{path[0][0][1000]}/{basename[0][0]}.tmp2",
                       extras=["{FILE_PART[0][0]}"])
test_pipeline1.combinations(task_func=check_combinations2_task,
                            input=generate_initial_files1,      # gen_task1
                            filter=formatter(".*/(?P<FILE_PART>.+).tmp1$"),
                            tuple_size=2,
                            output="{path[0][0]}/{FILE_PART[0][0]}.{basename[1][0]}.tmp2",
                            extras=["{basename[0][0][0]}{basename[1][0][0]}",       # extra: prefices
                                    # extra: path for 2nd input, 1st file
                                    "{subpath[0][0][0]}",
                                    "{subdir[0][0][0]}"])
test_pipeline1.merge(task_func=check_combinations2_merged_task,
                     input=check_combinations2_task,
                     output=tempdir + "/merged.results")
test_pipeline1.combinations(task_func=check_combinations3_task,
                            input=output_from("WOWWWEEE"),
                            filter=formatter(".*/(?P<FILE_PART>.+).tmp1$"),
                            tuple_size=3,
                            output="{path[0][0]}/{FILE_PART[0][0]}.{basename[1][0]}.{basename[2][0]}.tmp2",
                            extras=["{basename[0][0][0]}{basename[1][0][0]}{basename[2][0][0]}",       # extra: prefices
                                    # extra: path for 2nd input, 1st file
                                    "{subpath[0][0][0]}",
                                    "{subdir[0][0][0]}"])

test_pipeline1.merge(check_combinations3_merged_task,
                     check_combinations3_task, tempdir + "/merged.results")
test_pipeline1.permutations(task_func=check_permutations2_task,
                            input=output_from("WOWWWEEE"),
                            filter=formatter(".*/(?P<FILE_PART>.+).tmp1$"),
                            tuple_size=2,
                            output="{path[0][0]}/{FILE_PART[0][0]}.{basename[1][0]}.tmp2",
                            extras=["{basename[0][0][0]}{basename[1][0][0]}",       # extra: prefices
                                    # extra: path for 2nd input, 1st file
                                    "{subpath[0][0][0]}",
                                    "{subdir[0][0][0]}"])

test_pipeline2.merge(check_permutations2_merged_task,
                     check_permutations2_task, tempdir + "/merged.results")


test_pipeline2.permutations(task_func=check_permutations3_task,
                            input=output_from("WOWWWEEE"),
                            filter=formatter(".*/(?P<FILE_PART>.+).tmp1$"),
                            tuple_size=3,
                            output="{path[0][0]}/{FILE_PART[0][0]}.{basename[1][0]}.{basename[2][0]}.tmp2",
                            extras=["{basename[0][0][0]}{basename[1][0][0]}{basename[2][0][0]}",       # extra: prefices
                                    # extra: path for 2nd input, 1st file
                                    "{subpath[0][0][0]}",
                                    "{subdir[0][0][0]}"])
test_pipeline2.merge(check_permutations3_merged_task,
                     check_permutations3_task, tempdir + "/merged.results")
test_pipeline2.combinations_with_replacement(check_combinations_with_replacement2_task,
                                             input=output_from("WOWWWEEE"),
                                             filter=formatter(
                                                 ".*/(?P<FILE_PART>.+).tmp1$"),
                                             tuple_size=2,
                                             output="{path[0][0]}/{FILE_PART[0][0]}.{basename[1][0]}.tmp2",
                                             extras=["{basename[0][0][0]}{basename[1][0][0]}",       # extra: prefices
                                                     # extra: path for 2nd input, 1st file
                                                     "{subpath[0][0][0]}",
                                                     "{subdir[0][0][0]}"])
test_pipeline2.merge(check_combinations_with_replacement2_merged_task,
                     check_combinations_with_replacement2_task, tempdir + "/merged.results")
test_pipeline2.combinations_with_replacement(task_func=check_combinations_with_replacement3_task,
                                             input=output_from("WOWWWEEE"),
                                             filter=formatter(
                                                 ".*/(?P<FILE_PART>.+).tmp1$"),
                                             tuple_size=3,
                                             output="{path[0][0]}/{FILE_PART[0][0]}.{basename[1][0]}.{basename[2][0]}.tmp2",
                                             extras=["{basename[0][0][0]}{basename[1][0][0]}{basename[2][0][0]}",       # extra: prefices
                                                     # extra: path for 2nd input, 1st file
                                                     "{subpath[0][0][0]}",
                                                     "{subdir[0][0][0]}"])
test_pipeline2.merge(check_combinations_with_replacement3_merged_task,
                     check_combinations_with_replacement3_task, tempdir + "/merged.results")


class TestCombinatorics(unittest.TestCase):
    def setUp(self):
        try:
            os.mkdir(tempdir)
        except OSError:
            pass

    # ___________________________________________________________________________
    #
    #   test product() pipeline_printout and pipeline_run
    # ___________________________________________________________________________
    def test_product_printout(self):
        """Input file exists, output doesn't exist"""
        cleanup_tmpdir()
        s = StringIO()
        test_pipeline2.printout(
            s, [check_product_merged_task], verbose=5, wrap_width=10000)
        self.assertTrue(re.search('Job needs update:.*Missing files.*'
                                  '\[.*{tempdir}/a_name.tmp1, '
                                  '.*{tempdir}/e_name.tmp1, '
                                  '.*{tempdir}/h_name.tmp1, '
                                  '.*{tempdir}/a_name.e_name.h_name.tmp2\]'.format(tempdir=tempdir), s.getvalue(), re.DOTALL))

    def test_product_run(self):
        """Run product"""
        # output is up to date, but function body changed (e.g., source different)
        cleanup_tmpdir()
        test_pipeline2.run([check_product_merged_task], verbose=0,
                           multiprocess=100, one_second_per_job=one_second_per_job)
        with open(tempdir + "/merged.results") as oo:
            self.assertEqual(oo.read(),
                             "aeg,aeh,afg,afh,beg,beh,bfg,bfh,ceg,ceh,cfg,cfh,deg,deh,dfg,dfh,")

    # ___________________________________________________________________________
    #
    #   test product() pipeline_printout diagnostic error messsages
    #
    #       require verbose >= 3 or an empty jobs list
    # ___________________________________________________________________________
    def test_product_misspelt_capture_error(self):
        """Misspelt named capture group
            Requires verbose >= 3 or an empty jobs list
        """
        cleanup_tmpdir()

        s = StringIO()
        test_pipeline2.printout(
            s, [check_product_misspelt_capture_error_task], verbose=3, wrap_width=10000)
        self.assertIn("Warning: Input substitution failed:", s.getvalue())
        self.assertIn("Unmatched field {FILEPART}", s.getvalue())

    def test_product_out_of_range_formatter_ref_error(self):
        """
        {path[2][0]} when len(path) == 1
            Requires verbose >= 3 or an empty jobs list
        """
        cleanup_tmpdir()

        s = StringIO()
        test_pipeline2.printout(
            s, [check_product_out_of_range_formatter_ref_error_task], verbose=3, wrap_width=10000)
        self.assertIn("Warning: Input substitution failed:", s.getvalue())
        self.assertIn("Unmatched field {2}", s.getvalue())

    def test_product_formatter_ref_index_error(self):
        """
        {path[0][0][1000} when len of the path string len(path[0][0]) < 1000
            Requires verbose >= 3 or an empty jobs list
        """
        cleanup_tmpdir()

        s = StringIO()
        test_pipeline2.printout(
            s, [check_product_formatter_ref_index_error_task], verbose=3, wrap_width=10000)
        self.assertIn("Warning: Input substitution failed:", s.getvalue())
        self.assertIn(
            "Unmatched field {string index out of range}", s.getvalue())
        # print s.getvalue()

    # ___________________________________________________________________________
    #
    #   test combinations() pipeline_printout and pipeline_run
    # ___________________________________________________________________________

    def test_combinations2_printout(self):
        """Input file exists, output doesn't exist"""
        cleanup_tmpdir()

        s = StringIO()
        test_pipeline1.printout(
            s, [check_combinations2_merged_task], verbose=5, wrap_width=10000)
        self.assertTrue(re.search('Job needs update:.*Missing files.*'
                                  '\[.*{tempdir}/a_name.tmp1, '
                                  '.*{tempdir}/b_name.tmp1, '
                                  '.*{tempdir}/a_name.b_name.tmp2\]'.format(tempdir=tempdir), s.getvalue(), re.DOTALL))

    def test_combinations2_run(self):
        """Run product"""
        # output is up to date, but function body changed (e.g., source different)
        cleanup_tmpdir()
        test_pipeline2.run([check_combinations2_merged_task], verbose=0,
                           multiprocess=100, one_second_per_job=one_second_per_job)
        with open(tempdir + "/merged.results") as oo:
            self.assertEqual(oo.read(),
                             'ab,ac,ad,bc,bd,cd,')

    # ___________________________________________________________________________
    #
    #   test combinations() pipeline_printout and pipeline_run
    # ___________________________________________________________________________
    def test_combinations3_printout(self):
        """Input file exists, output doesn't exist"""
        cleanup_tmpdir()

        s = StringIO()
        test_pipeline2.printout(
            s, [check_combinations3_merged_task], verbose=5, wrap_width=10000)
        self.assertTrue(re.search(
            '\[.*{tempdir}/a_name.tmp1, '
            '.*{tempdir}/b_name.tmp1, '
            '.*{tempdir}/c_name.tmp1, '
            '.*{tempdir}/a_name.b_name.c_name.tmp2\]'.format(tempdir=tempdir), s.getvalue()))

    def test_combinations3_run(self):
        """Run product"""
        # output is up to date, but function body changed (e.g., source different)
        cleanup_tmpdir()
        test_pipeline2.run([check_combinations3_merged_task], verbose=0,
                           multiprocess=100, one_second_per_job=one_second_per_job)
        with open(tempdir + "/merged.results") as oo:
            self.assertEqual(oo.read(),
                             "abc,abd,acd,bcd,")

    # ___________________________________________________________________________
    #
    #   test permutations() pipeline_printout and pipeline_run
    # ___________________________________________________________________________

    def test_permutations2_printout(self):
        """Input file exists, output doesn't exist"""
        cleanup_tmpdir()

        s = StringIO()
        test_pipeline2.printout(
            s, [check_permutations2_merged_task], verbose=5, wrap_width=10000)
        self.assertTrue(re.search('\[.*{tempdir}/a_name.tmp1, '
                                  '.*{tempdir}/b_name.tmp1, '
                                  '.*{tempdir}/a_name.b_name.tmp2\]'.format(tempdir=tempdir), s.getvalue()))

    def test_permutations2_run(self):
        """Run product"""
        # output is up to date, but function body changed (e.g., source different)
        cleanup_tmpdir()
        test_pipeline2.run([check_permutations2_merged_task], verbose=0,
                           multiprocess=100, one_second_per_job=one_second_per_job)
        with open(tempdir + "/merged.results") as oo:
            self.assertEqual(oo.read(),
                             "ab,ac,ad,ba,bc,bd,ca,cb,cd,da,db,dc,")

    # ___________________________________________________________________________
    #
    #   test permutations() pipeline_printout and pipeline_run
    # ___________________________________________________________________________
    def test_permutations3_printout(self):
        """Input file exists, output doesn't exist"""
        cleanup_tmpdir()

        s = StringIO()
        test_pipeline2.printout(
            s, [check_permutations3_merged_task], verbose=5, wrap_width=10000)
        self.assertTrue(re.search('\[.*{tempdir}/a_name.tmp1, '
                                  '.*{tempdir}/b_name.tmp1, '
                                  '.*{tempdir}/c_name.tmp1, '
                                  '.*{tempdir}/a_name.b_name.c_name.tmp2\]'.format(tempdir=tempdir), s.getvalue()))

    def test_permutations3_run(self):
        """Run product"""
        # output is up to date, but function body changed (e.g., source different)
        cleanup_tmpdir()
        test_pipeline2.run([check_permutations3_merged_task], verbose=0,
                           multiprocess=100, one_second_per_job=one_second_per_job)
        with open(tempdir + "/merged.results") as oo:
            self.assertEqual(oo.read(),
                             'abc,abd,acb,acd,adb,adc,bac,bad,bca,bcd,bda,bdc,cab,cad,cba,cbd,cda,cdb,dab,dac,dba,dbc,dca,dcb,')

    # ___________________________________________________________________________
    #
    #   test combinations_with_replacement() pipeline_printout and pipeline_run
    # ___________________________________________________________________________

    def test_combinations_with_replacement2_printout(self):
        """Input file exists, output doesn't exist"""
        cleanup_tmpdir()

        s = StringIO()
        test_pipeline2.printout(
            s, [check_combinations_with_replacement2_merged_task], verbose=5, wrap_width=10000)
        self.assertTrue(re.search('\[.*{tempdir}/a_name.tmp1, '
                                  '.*{tempdir}/b_name.tmp1, '
                                  '.*{tempdir}/a_name.b_name.tmp2\]'.format(tempdir=tempdir), s.getvalue()))

    def test_combinations_with_replacement2_run(self):
        """Run product"""
        # output is up to date, but function body changed (e.g., source different)
        cleanup_tmpdir()
        test_pipeline2.run([check_combinations_with_replacement2_merged_task],
                           verbose=0, multiprocess=100, one_second_per_job=one_second_per_job)
        with open(tempdir + "/merged.results") as oo:
            self.assertEqual(oo.read(),
                             "aa,ab,ac,ad,bb,bc,bd,cc,cd,dd,")

    # ___________________________________________________________________________
    #
    #   test combinations_with_replacement() pipeline_printout and pipeline_run
    # ___________________________________________________________________________
    def test_combinations_with_replacement3_printout(self):
        """Input file exists, output doesn't exist"""
        cleanup_tmpdir()

        s = StringIO()
        test_pipeline2.printout(
            s, [check_combinations_with_replacement3_merged_task], verbose=5, wrap_width=10000)
        self.assertTrue(re.search('\[.*{tempdir}/a_name.tmp1, '
                                  '.*{tempdir}/b_name.tmp1, '
                                  '.*{tempdir}/c_name.tmp1, '
                                  '.*{tempdir}/a_name.b_name.c_name.tmp2\]'.format(tempdir=tempdir), s.getvalue()))

    def test_combinations_with_replacement3_run(self):
        """Run product"""
        # output is up to date, but function body changed (e.g., source different)
        cleanup_tmpdir()
        test_pipeline2.run([check_combinations_with_replacement3_merged_task],
                           verbose=0, multiprocess=100, one_second_per_job=one_second_per_job)
        with open(tempdir + "/merged.results") as oo:
            self.assertEqual(oo.read(),
                             'aaa,aab,aac,aad,abb,abc,abd,acc,acd,add,bbb,bbc,bbd,bcc,bcd,bdd,ccc,ccd,cdd,ddd,')

    # ___________________________________________________________________________
    #
    #   cleanup
    # ___________________________________________________________________________

    def tearDown(self):
        shutil.rmtree(tempdir)


#
#   Necessary to protect the "entry point" of the program under windows.
#       see: http://docs.python.org/library/multiprocessing.html#multiprocessing-programming
#
if __name__ == '__main__':
    #pipeline_printout(sys.stdout, [check_product_task], verbose = 5)
    unittest.main()



################################################################################
#
#
#   combinatorics.py
#
#   Copyright (c) 2013 Leo Goodstadt
#
#   Permission is hereby granted, free of charge, to any person obtaining a copy
#   of this software and associated documentation files (the "Software"), to deal
#   in the Software without restriction, including without limitation the rights
#   to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#   copies of the Software, and to permit persons to whom the Software is
#   furnished to do so, subject to the following conditions:
#
#   The above copyright notice and this permission notice shall be included in
#   all copies or substantial portions of the Software.
#
#   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#   AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#   LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#   OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#   THE SOFTWARE.
#################################################################################
"""

********************************************
:mod:`ruffus.combinatorics` -- Overview
********************************************

.. moduleauthor:: Leo Goodstadt <ruffus@llew.org.uk>


    #
    #   @product
    #
    from ruffus import *

    @product(
        "/a_directory/dataset1.*.bam",
        formatter( ".*/dataset1\.(?P<ID>\d+).bam" ),
        "/a_different/directory/dataset2.*.bam",
        formatter(".species", ".*/dataset2\.(?P<ID>\d+).bam" ),
        "{path[0][0]}/{base_name[0][0]}.{base_name[0][0]}.out",
        "{path[0][0]}",       # extra: path for 1st input, 1st file
        "{path[1][0]}",       # extra: path for 2nd input, 1st file
        "{basename[0][1]}",   # extra: file name for 1st input, 2nd file
        "{ID[1][1]}",         # extra: regular expression named capture group for 2nd input, 2nd file
        )
    def task1( infiles, outfile,
                input_1__path,
                input_2__path,
                input_1__2nd_file_name,
                input_2__3rd_file_match
                ):
        print infiles
        print outfile
        print input_1__path
        print input_2__path
        print input_1__2nd_file_name
        print input_2__3rd_file_match
"""

from .task import task_decorator


#
#   Combinatoric generators:
#
class product(task_decorator):
    pass


class permutations(task_decorator):
    pass


class combinations(task_decorator):
    pass


class combinations_with_replacement(task_decorator):
    pass



#!/usr/bin/env python
from __future__ import print_function
import shutil
import unittest
import glob
import random
from ruffus import follows, split, mkdir, files, transform, suffix, posttask, touch_file, merge, Pipeline
import sys
import os



NUMBER_OF_RANDOMS = 10000
CHUNK_SIZE = 1000


tempdir = os.path.relpath(os.path.abspath(os.path.splitext(__file__)[0])) + "/"

# add grandparent to search path for testing
grandparent_dir = os.path.abspath(
    os.path.join(os.path.dirname(__file__), "..", ".."))
sys.path.insert(0, grandparent_dir)

# module name = script name without extension
module_name = os.path.splitext(os.path.basename(__file__))[0]


# ---------------------------------------------------------------
#
#   Create random numbers
#
@follows(mkdir(tempdir))
@files(None, tempdir + "random_numbers.list")
def create_random_numbers(input_file_name, output_file_name):
    f = open(output_file_name, "w")
    for i in range(NUMBER_OF_RANDOMS):
        f.write("%g\n" % (random.random() * 100.0))
    f.close()

# ---------------------------------------------------------------
#
#   Split initial file
#


@follows(create_random_numbers)
@split(tempdir + "random_numbers.list", tempdir + "*.chunks")
def step_4_split_numbers_into_chunks(input_file_name, output_files):
    """
        Splits random numbers file into XXX files of CHUNK_SIZE each
    """
    #
    #   clean up files from previous runs
    #
    for f in glob.glob("*.chunks"):
        os.unlink(f)
    #
    #   create new file every CHUNK_SIZE lines and
    #       copy each line into current file
    #
    output_file = None
    cnt_files = 0
    with open(input_file_name) as ii:
        for i, line in enumerate(ii):
            if i % CHUNK_SIZE == 0:
                cnt_files += 1
                if output_file:
                    output_file.close()
                output_file = open(tempdir + "%d.chunks" % cnt_files, "w")
            output_file.write(line)
    if output_file:
        output_file.close()

# ---------------------------------------------------------------
#
#   Calculate sum and sum of squares for each chunk file
#


@transform(step_4_split_numbers_into_chunks, suffix(".chunks"), ".sums")
def step_5_calculate_sum_of_squares(input_file_name, output_file_name):
    with open(output_file_name,  "w") as oo:
        sum_squared, sum = [0.0, 0.0]
        cnt_values = 0
        with open(input_file_name) as ii:
            for line in ii:
                cnt_values += 1
                val = float(line.rstrip())
                sum_squared += val * val
                sum += val
        oo.write("%s\n%s\n%d\n" % (repr(sum_squared), repr(sum), cnt_values))


def print_hooray_again():
    print("     hooray again")


def print_whoppee_again():
    print("     whoppee again")


# ---------------------------------------------------------------
#
#   Calculate sum and sum of squares for each chunk
#
@posttask(lambda: sys.stdout.write("     hooray\n"))
@posttask(print_hooray_again, print_whoppee_again, touch_file(os.path.join(tempdir, "done")))
@merge(step_5_calculate_sum_of_squares, os.path.join(tempdir, "variance.result"))
def step_6_calculate_variance(input_file_names, output_file_name):
    """
    Calculate variance naively
    """
    output = open(output_file_name,  "w")
    #
    #   initialise variables
    #
    all_sum_squared = 0.0
    all_sum = 0.0
    all_cnt_values = 0.0
    #
    # added up all the sum_squared, and sum and cnt_values from all the chunks
    #
    for input_file_name in input_file_names:
        with open(input_file_name) as ii:
            sum_squared, sum, cnt_values = list(map(float, ii.readlines()))
        all_sum_squared += sum_squared
        all_sum += sum
        all_cnt_values += cnt_values
    all_mean = all_sum / all_cnt_values
    variance = (all_sum_squared - all_sum * all_mean)/(all_cnt_values)
    #
    #   print output
    #
    print(variance, file=output)
    output.close()


try:
    from StringIO import StringIO
except:
    from io import StringIO


class Test_ruffus(unittest.TestCase):
    def setUp(self):
        try:
            shutil.rmtree(tempdir)
        except:
            pass

    def tearDown(self):
        try:
            shutil.rmtree(tempdir)
            pass
        except:
            pass

    def atest_ruffus(self):
        pipeline_run(multiprocess=50, verbose=0, pipeline="main")
        output_file = os.path.join(tempdir, "variance.result")
        if not os.path.exists(output_file):
            raise Exception("Missing %s" % output_file)

    def test_newstyle_ruffus(self):
        test_pipeline = Pipeline("test")

        test_pipeline.files(create_random_numbers, None, tempdir + "random_numbers.list")\
            .follows(mkdir(tempdir))

        test_pipeline.split(task_func=step_4_split_numbers_into_chunks,
                            input=tempdir + "random_numbers.list",
                            output=tempdir + "*.chunks")\
            .follows(create_random_numbers)

        test_pipeline.transform(task_func=step_5_calculate_sum_of_squares,
                                input=step_4_split_numbers_into_chunks,
                                filter=suffix(".chunks"),
                                output=".sums")

        test_pipeline.merge(task_func=step_6_calculate_variance, input=step_5_calculate_sum_of_squares, output=os.path.join(tempdir, "variance.result"))\
            .posttask(lambda: sys.stdout.write("     hooray\n"))\
            .posttask(print_hooray_again, print_whoppee_again, touch_file(os.path.join(tempdir, "done")))

        test_pipeline.run(multiprocess=50, verbose=0)
        output_file = os.path.join(tempdir, "variance.result")
        if not os.path.exists(output_file):
            raise Exception("Missing %s" % output_file)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
from __future__ import print_function
import re
import time
import shutil
import unittest
from ruffus.ruffus_exceptions import RethrownJobError
from ruffus.ruffus_utility import RUFFUS_HISTORY_FILE, CHECKSUM_FILE_TIMESTAMPS, \
    CHECKSUM_HISTORY_TIMESTAMPS, CHECKSUM_FUNCTIONS, CHECKSUM_FUNCTIONS_AND_PARAMS
from ruffus.ruffus_utility import get_default_history_file_name
from ruffus import pipeline_run, pipeline_printout, suffix, transform, split, merge, dbdict, Pipeline
import sys

"""

    test_job_completion_transform.py

        test several cases where the dbdict should be updated

"""

import os
tempdir = os.path.relpath(os.path.abspath(os.path.splitext(__file__)[0])) + "/"

# add grandparent to search path for testing
grandparent_dir = os.path.abspath(
    os.path.join(os.path.dirname(__file__), "..", ".."))
sys.path.insert(0, grandparent_dir)

# module name = script name without extension
module_name = os.path.splitext(os.path.basename(__file__))[0]


# funky code to import by file name
parent_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), ".."))


# ___________________________________________________________________________
#
#   imports
# ___________________________________________________________________________


try:
    from StringIO import StringIO
except:
    from io import StringIO


possible_chksms = list(range(CHECKSUM_FUNCTIONS_AND_PARAMS + 1))
tempdir = 'tmp_test_job_completion/'
input_file = os.path.join(tempdir, 'input.txt')
transform1_out = input_file.replace('.txt', '.output')
split1_outputs = [os.path.join(tempdir, 'split.out1.txt'),
                  os.path.join(tempdir, 'split.out2.txt')]
merge2_output = os.path.join(tempdir, 'merged.out')

runtime_data = []


@transform(input_file, suffix('.txt'), '.output', runtime_data)
def transform1(in_name, out_name, how_many):
    with open(out_name, 'w') as outfile:
        with open(in_name) as ii:
            outfile.write(ii.read())


@transform(input_file, suffix('.txt'), '.output', runtime_data)
def transform_raise_error(in_name, out_name, how_many):
    # raise an error unless runtime_data has 'okay' in it
    with open(out_name, 'w') as outfile:
        with open(in_name) as ii:
            outfile.write(ii.read())
    if 'okay' not in how_many:
        raise RuntimeError("'okay' wasn't in runtime_data!")


@split(input_file, split1_outputs)
def split1(in_name, out_names):
    for n in out_names:
        with open(n, 'w') as outfile:
            with open(in_name) as ii:
                outfile.write(ii.read() + '\n')


@merge(split1, merge2_output)
def merge2(in_names, out_name):
    with open(out_name, 'w') as outfile:
        for n in in_names:
            with open(n) as ii:
                outfile.write(ii.read() + '\n')


# CHECKSUM_FILE_TIMESTAMPS      = 0     # only rerun when the file timestamps are out of date (classic mode)
# CHECKSUM_HISTORY_TIMESTAMPS   = 1     # also rerun when the history shows a job as being out of date
# CHECKSUM_FUNCTIONS            = 2     # also rerun when function body has changed
# CHECKSUM_FUNCTIONS_AND_PARAMS = 3     # also rerun when function parameters have changed


def cleanup_tmpdir():
    os.system('rm -f %s %s' %
              (os.path.join(tempdir, '*'), get_default_history_file_name()))


count_pipelines = 0


class TestJobCompletion(unittest.TestCase):
    def setUp(self):
        try:
            os.mkdir(tempdir)
        except OSError:
            pass

    def create_pipeline(self):
        """
        Create new pipeline on the fly without using decorators
        """
        global count_pipelines
        count_pipelines = count_pipelines + 1
        test_pipeline = Pipeline("test %d" % count_pipelines)

        test_pipeline.transform(task_func=transform1,
                                input=input_file,
                                filter=suffix('.txt'),
                                output='.output',
                                extras=[runtime_data])

        test_pipeline.transform(task_func=transform_raise_error,
                                input=input_file,
                                filter=suffix('.txt'),
                                output='.output',
                                extras=[runtime_data])

        test_pipeline.split(task_func=split1,
                            input=input_file,
                            output=split1_outputs)

        test_pipeline.merge(task_func=merge2,
                            input=split1,
                            output=merge2_output)
        return test_pipeline

    def test_output_doesnt_exist(self):
        """Input file exists, output doesn't exist"""
        # output doesn't exist-- should run for all levels
        # create a new input file
        cleanup_tmpdir()
        with open(input_file, 'w') as outfile:
            outfile.write('testme')

        for chksm in possible_chksms:
            s = StringIO()
            pipeline_printout(s, [transform1], verbose=6,
                              checksum_level=chksm, pipeline="main")
            self.assertTrue(re.search(
                r'Job needs update:.*Missing file.*\[tmp_test_job_completion/input.output\]', s.getvalue(), re.DOTALL))

    def test_output_out_of_date(self):
        """Input file exists, output out of date"""
        # output exists but is out of date-- should run for all levels
        cleanup_tmpdir()
        with open(transform1_out, 'w') as outfile:
            outfile.write('testme')
        time.sleep(0.1)
        with open(input_file, 'w') as outfile:
            outfile.write('testme')

        for chksm in possible_chksms:
            s = StringIO()
            pipeline_printout(s, [transform1], verbose=6,
                              checksum_level=chksm, pipeline="main")
            self.assertIn('Job needs update:', s.getvalue())
            if chksm == CHECKSUM_FILE_TIMESTAMPS:
                self.assertIn('Input files:', s.getvalue())
                self.assertIn('Output files:', s.getvalue())
            else:
                self.assertIn('left over from a failed run?', s.getvalue())

    def test_output_timestamp_okay(self):
        """Input file exists, output timestamp up to date"""
        # output exists and timestamp is up to date-- not run for lvl 0, run for all others
        cleanup_tmpdir()
        with open(input_file, 'w') as outfile:
            outfile.write('testme')
        time.sleep(0.1)
        with open(transform1_out, 'w') as outfile:
            outfile.write('testme')

        for chksm in possible_chksms:
            s = StringIO()
            pipeline_printout(s, [transform1], verbose=6,
                              checksum_level=chksm, pipeline="main")
            if chksm == CHECKSUM_FILE_TIMESTAMPS:
                #self.assertIn('Job up-to-date', s.getvalue())
                pass
            else:
                self.assertIn('Job needs update:', s.getvalue())
                self.assertIn('left over from a failed run?',
                              s.getvalue())

    def test_output_up_to_date(self):
        """Input file exists, output up to date"""
        # output is up to date-- not run for any levels
        cleanup_tmpdir()
        with open(input_file, 'w') as outfile:
            outfile.write('testme')
        pipeline_run([transform1], verbose=0,
                     checksum_level=CHECKSUM_HISTORY_TIMESTAMPS, pipeline="main")

        for chksm in possible_chksms:
            s = StringIO()
            pipeline_printout(s, [transform1], verbose=6,
                              checksum_level=chksm, pipeline="main")
            #self.assertIn('Job up-to-date', s.getvalue())
            pass

    def test_output_up_to_date_func_changed(self):
        """Input file exists, output up to date, function body changed"""
        # output is up to date, but function body changed (e.g., source different)
        cleanup_tmpdir()
        with open(input_file, 'w') as outfile:
            outfile.write('testme')
        pipeline_run([transform1], verbose=0,
                     checksum_level=CHECKSUM_HISTORY_TIMESTAMPS, pipeline="main")
        if sys.hexversion >= 0x03000000:
            transform1.__code__ = split1.__code__  # simulate source change
        else:
            transform1.func_code = split1.func_code  # simulate source change

        for chksm in possible_chksms:
            s = StringIO()
            pipeline_printout(s, [transform1], verbose=6,
                              checksum_level=chksm, pipeline="main")
            if chksm >= CHECKSUM_FUNCTIONS:
                self.assertIn('Job needs update:', s.getvalue())
                self.assertIn('Pipeline function has changed',
                              s.getvalue())
            else:
                #self.assertIn('Job up-to-date', s.getvalue())
                pass

    def test_output_up_to_date_func_changed(self):
        """Input file exists, output up to date, function body changed"""
        # output is up to date, but function body changed (e.g., source different)
        cleanup_tmpdir()
        with open(input_file, 'w') as outfile:
            outfile.write('testme')
        pipeline_run([transform1], verbose=0,
                     checksum_level=CHECKSUM_HISTORY_TIMESTAMPS, pipeline="main")
        # simulate source change
        if sys.hexversion >= 0x03000000:
            split1.__code__, transform1.__code__ = transform1.__code__, split1.__code__
        else:
            split1.func_code, transform1.func_code = transform1.func_code, split1.func_code

        for chksm in possible_chksms:
            s = StringIO()
            pipeline_printout(s, [transform1], verbose=6,
                              checksum_level=chksm, pipeline="main")
            if chksm >= CHECKSUM_FUNCTIONS:
                self.assertIn('Job needs update:', s.getvalue())
                self.assertIn('Pipeline function has changed',
                              s.getvalue())
            else:
                #self.assertIn('Job up-to-date', s.getvalue())
                pass
        # clean up our function-changing mess!
        if sys.hexversion >= 0x03000000:
            split1.__code__, transform1.__code__ = transform1.__code__, split1.__code__
        else:
            split1.func_code, transform1.func_code = transform1.func_code, split1.func_code

    def test_output_up_to_date_param_changed(self):
        """Input file exists, output up to date, parameter to function changed"""
        # output is up to date, but function body changed (e.g., source different)
        cleanup_tmpdir()
        with open(input_file, 'w') as outfile:
            outfile.write('testme')
        pipeline_run([transform1], verbose=0,
                     checksum_level=CHECKSUM_HISTORY_TIMESTAMPS, pipeline="main")
        runtime_data.append('different')  # simulate change to config file

        for chksm in possible_chksms:
            s = StringIO()
            pipeline_printout(s, [transform1], verbose=6,
                              checksum_level=chksm, pipeline="main")
            if chksm >= CHECKSUM_FUNCTIONS_AND_PARAMS:
                self.assertIn('Job needs update:', s.getvalue())
                self.assertIn('Pipeline parameters have changed',
                              s.getvalue())
            else:
                #self.assertIn('Job up-to-date', s.getvalue())
                pass

    def test_split_output(self):
        """test multiple-output checksums"""
        # outputs out of date
        cleanup_tmpdir()
        with open(input_file, 'w') as outfile:
            outfile.write('testme')
        pipeline_run([split1], verbose=0,
                     checksum_level=CHECKSUM_HISTORY_TIMESTAMPS, pipeline="main")
        time.sleep(.5)
        with open(input_file, 'w') as outfile:
            outfile.write('testme')

        for chksm in possible_chksms:
            s = StringIO()
            pipeline_printout(s, [split1], verbose=6,
                              checksum_level=chksm, pipeline="main")
            self.assertIn('Job needs update:', s.getvalue())

        # all outputs incorrectly generated
        cleanup_tmpdir()
        with open(input_file, 'w') as outfile:
            outfile.write('testme')
        time.sleep(.5)
        for f in split1_outputs:
            with open(f, 'w') as outfile:
                outfile.write('testme')
        for chksm in possible_chksms:
            s = StringIO()
            pipeline_printout(s, [split1], verbose=6,
                              checksum_level=chksm, pipeline="main")
            if chksm >= CHECKSUM_HISTORY_TIMESTAMPS:
                self.assertIn('Job needs update:', s.getvalue())
                self.assertIn('left over from a failed run?',
                              s.getvalue())
            else:
                #self.assertIn('Job up-to-date', s.getvalue())
                pass

        # one output incorrectly generated
        cleanup_tmpdir()
        with open(input_file, 'w') as outfile:
            outfile.write('testme')
        pipeline_run([split1], verbose=0,
                     checksum_level=CHECKSUM_HISTORY_TIMESTAMPS, pipeline="main")
        job_history = dbdict.open(
            get_default_history_file_name(), picklevalues=True)
        del job_history[os.path.relpath(split1_outputs[0])]

        for chksm in possible_chksms:
            s = StringIO()
            pipeline_printout(s, [split1], verbose=6,
                              checksum_level=chksm, pipeline="main")
            if chksm >= CHECKSUM_HISTORY_TIMESTAMPS:
                self.assertIn('Job needs update:', s.getvalue())
                self.assertIn('left over from a failed run?',
                              s.getvalue())
            else:
                #self.assertIn('Job up-to-date', s.getvalue())
                pass

    def test_merge_output(self):
        """test multiple-input checksums"""
        # one output incorrectly generated
        cleanup_tmpdir()
        with open(input_file, 'w') as outfile:
            outfile.write('testme')
        pipeline_run([split1], verbose=0,
                     checksum_level=CHECKSUM_HISTORY_TIMESTAMPS, pipeline="main")
        job_history = dbdict.open(
            get_default_history_file_name(), picklevalues=True)
        del job_history[os.path.relpath(split1_outputs[0])]

        for chksm in possible_chksms:
            s = StringIO()
            pipeline_printout(s, [merge2], verbose=6,
                              checksum_level=chksm, pipeline="main")
            if chksm >= CHECKSUM_HISTORY_TIMESTAMPS:
                self.assertIn('Job needs update:', s.getvalue())
                self.assertIn('left over from a failed run?', s.getvalue())
            else:
                #self.assertIn('Job up-to-date', s.getvalue())
                pass

        # make sure the jobs run fine
        cleanup_tmpdir()
        with open(input_file, 'w') as outfile:
            outfile.write('testme')
        pipeline_run([merge2], verbose=0,
                     checksum_level=CHECKSUM_HISTORY_TIMESTAMPS, pipeline="main")
        for chksm in possible_chksms:
            s = StringIO()
            pipeline_printout(s, [merge2], verbose=6,
                              checksum_level=chksm, pipeline="main")
            #self.assertIn('Job up-to-date', s.getvalue())
            self.assertNotIn('Job needs update:', s.getvalue())
            self.assertNotIn('left over from a failed run?', s.getvalue())

    def test_newstyle_output_doesnt_exist(self):
        """Input file exists, output doesn't exist"""
        # output doesn't exist-- should run for all levels
        # create a new input file
        cleanup_tmpdir()
        with open(input_file, 'w') as outfile:
            outfile.write('testme')

        for chksm in possible_chksms:
            s = StringIO()
            self.create_pipeline().printout(
                s, [transform1], verbose=6, checksum_level=chksm)
            self.assertTrue(re.search(
                r'Job needs update:.*Missing file.*\[tmp_test_job_completion/input.output\]', s.getvalue(), re.DOTALL))

    def test_newstyle_output_out_of_date(self):
        """Input file exists, output out of date"""
        # output exists but is out of date-- should run for all levels
        cleanup_tmpdir()
        with open(transform1_out, 'w') as outfile:
            outfile.write('testme')
        time.sleep(0.1)
        with open(input_file, 'w') as outfile:
            outfile.write('testme')

        for chksm in possible_chksms:
            s = StringIO()
            self.create_pipeline().printout(
                s, [transform1], verbose=6, checksum_level=chksm)
            self.assertIn('Job needs update:', s.getvalue())
            if chksm == CHECKSUM_FILE_TIMESTAMPS:
                self.assertIn('Input files:', s.getvalue())
                self.assertIn('Output files:', s.getvalue())
            else:
                self.assertIn('left over from a failed run?', s.getvalue())

    def test_newstyle_output_timestamp_okay(self):
        """Input file exists, output timestamp up to date"""
        # output exists and timestamp is up to date-- not run for lvl 0, run for all others
        cleanup_tmpdir()
        with open(input_file, 'w') as outfile:
            outfile.write('testme')
        time.sleep(0.1)
        with open(transform1_out, 'w') as outfile:
            outfile.write('testme')

        for chksm in possible_chksms:
            s = StringIO()
            self.create_pipeline().printout(
                s, [transform1], verbose=6, checksum_level=chksm)
            if chksm == CHECKSUM_FILE_TIMESTAMPS:
                #self.assertIn('Job up-to-date', s.getvalue())
                pass
            else:
                self.assertIn('Job needs update:', s.getvalue())
                self.assertIn('left over from a failed run?',
                              s.getvalue())

    def test_newstyle_output_up_to_date(self):
        """Input file exists, output up to date"""
        test_pipeline = self.create_pipeline()
        # output is up to date-- not run for any levels
        cleanup_tmpdir()
        with open(input_file, 'w') as outfile:
            outfile.write('testme')
        test_pipeline.run([transform1], verbose=0,
                          checksum_level=CHECKSUM_HISTORY_TIMESTAMPS)

        for chksm in possible_chksms:
            s = StringIO()
            test_pipeline.printout(
                s, [transform1], verbose=6, checksum_level=chksm)
            #self.assertIn('Job up-to-date', s.getvalue())
            pass

    def test_newstyle_output_up_to_date_func_changed(self):
        """Input file exists, output up to date, function body changed"""
        test_pipeline = self.create_pipeline()
        # output is up to date, but function body changed (e.g., source different)
        cleanup_tmpdir()
        with open(input_file, 'w') as outfile:
            outfile.write('testme')
        test_pipeline.run([transform1], verbose=0,
                          checksum_level=CHECKSUM_HISTORY_TIMESTAMPS)
        if sys.hexversion >= 0x03000000:
            transform1.__code__ = split1.__code__  # simulate source change
        else:
            transform1.func_code = split1.func_code  # simulate source change

        for chksm in possible_chksms:
            s = StringIO()
            test_pipeline.printout(
                s, [transform1], verbose=6, checksum_level=chksm)
            if chksm >= CHECKSUM_FUNCTIONS:
                self.assertIn('Job needs update:', s.getvalue())
                self.assertIn('Pipeline function has changed',
                              s.getvalue())
            else:
                #self.assertIn('Job up-to-date', s.getvalue())
                pass

    def test_newstyle_output_up_to_date_func_changed(self):
        """Input file exists, output up to date, function body changed"""
        test_pipeline = self.create_pipeline()
        # output is up to date, but function body changed (e.g., source different)
        cleanup_tmpdir()
        with open(input_file, 'w') as outfile:
            outfile.write('testme')
        test_pipeline.run([transform1], verbose=0,
                          checksum_level=CHECKSUM_HISTORY_TIMESTAMPS)
        # simulate source change
        if sys.hexversion >= 0x03000000:
            split1.__code__, transform1.__code__ = transform1.__code__, split1.__code__
        else:
            split1.func_code, transform1.func_code = transform1.func_code, split1.func_code

        for chksm in possible_chksms:
            s = StringIO()
            test_pipeline.printout(
                s, [transform1], verbose=6, checksum_level=chksm)
            if chksm >= CHECKSUM_FUNCTIONS:
                self.assertIn('Job needs update:', s.getvalue())
                self.assertIn('Pipeline function has changed',
                              s.getvalue())
            else:
                #self.assertIn('Job up-to-date', s.getvalue())
                pass
        # clean up our function-changing mess!
        if sys.hexversion >= 0x03000000:
            split1.__code__, transform1.__code__ = transform1.__code__, split1.__code__
        else:
            split1.func_code, transform1.func_code = transform1.func_code, split1.func_code

    def test_newstyle_output_up_to_date_param_changed(self):
        """Input file exists, output up to date, parameter to function changed"""
        test_pipeline = self.create_pipeline()
        # output is up to date, but function body changed (e.g., source different)
        cleanup_tmpdir()
        with open(input_file, 'w') as outfile:
            outfile.write('testme')
        test_pipeline.run([transform1], verbose=0,
                          checksum_level=CHECKSUM_HISTORY_TIMESTAMPS)
        runtime_data.append('different')  # simulate change to config file

        for chksm in possible_chksms:
            s = StringIO()
            test_pipeline.printout(
                s, [transform1], verbose=6, checksum_level=chksm)
            if chksm >= CHECKSUM_FUNCTIONS_AND_PARAMS:
                self.assertIn('Job needs update:', s.getvalue())
                self.assertIn('Pipeline parameters have changed',
                              s.getvalue())
            else:
                #self.assertIn('Job up-to-date', s.getvalue())
                pass

    def test_raises_error(self):
        """run a function that fails but creates output, then check what should run"""
        # output is up to date, but function body changed (e.g., source different)
        cleanup_tmpdir()
        with open(input_file, 'w') as outfile:
            outfile.write('testme')
        time.sleep(.5)
        del runtime_data[:]
        # poo. Shouldn't this be RuntimeError?
        with self.assertRaises(RethrownJobError):
            # generates output then fails
            pipeline_run([transform_raise_error], verbose=0,
                         checksum_level=CHECKSUM_HISTORY_TIMESTAMPS, pipeline="main")

        for chksm in possible_chksms:
            s = StringIO()
            pipeline_printout(s, [transform_raise_error],
                              verbose=6, checksum_level=chksm, pipeline="main")
            if chksm >= CHECKSUM_HISTORY_TIMESTAMPS:
                self.assertIn('Job needs update:', s.getvalue())
                self.assertIn('left over from a failed run?',
                              s.getvalue())
            else:
                #self.assertIn('Job up-to-date', s.getvalue())
                pass

    def test_newstyle_raises_error(self):
        """run a function that fails but creates output, then check what should run"""
        test_pipeline = self.create_pipeline()
        # output is up to date, but function body changed (e.g., source different)
        cleanup_tmpdir()
        with open(input_file, 'w') as outfile:
            outfile.write('testme')
        time.sleep(.5)
        del runtime_data[:]
        # poo. Shouldn't this be RuntimeError?
        with self.assertRaises(RethrownJobError):
            # generates output then fails
            test_pipeline.run([transform_raise_error], verbose=0,
                              checksum_level=CHECKSUM_HISTORY_TIMESTAMPS)
        for chksm in possible_chksms:
            s = StringIO()
            test_pipeline.printout(
                s, [transform_raise_error], verbose=6, checksum_level=chksm)
            if chksm >= CHECKSUM_HISTORY_TIMESTAMPS:
                self.assertIn('Job needs update:', s.getvalue())
                self.assertIn('left over from a failed run?',
                              s.getvalue())
            else:
                #self.assertIn('Job up-to-date', s.getvalue())
                pass

    def test_newstyle_split_output(self):
        """test multiple-output checksums"""
        test_pipeline = self.create_pipeline()
        # outputs out of date
        cleanup_tmpdir()
        with open(input_file, 'w') as outfile:
            outfile.write('testme')
        test_pipeline.run([split1], verbose=0,
                          checksum_level=CHECKSUM_HISTORY_TIMESTAMPS)
        time.sleep(.5)
        with open(input_file, 'w') as outfile:
            outfile.write('testme')

        for chksm in possible_chksms:
            s = StringIO()
            test_pipeline.printout(
                s, [split1], verbose=6, checksum_level=chksm)
            self.assertIn('Job needs update:', s.getvalue())

        # all outputs incorrectly generated
        cleanup_tmpdir()
        with open(input_file, 'w') as outfile:
            outfile.write('testme')
        time.sleep(.5)
        for f in split1_outputs:
            with open(f, 'w') as outfile:
                outfile.write('testme')
        for chksm in possible_chksms:
            s = StringIO()
            test_pipeline.printout(
                s, [split1], verbose=6, checksum_level=chksm)
            if chksm >= CHECKSUM_HISTORY_TIMESTAMPS:
                self.assertIn('Job needs update:', s.getvalue())
                self.assertIn('left over from a failed run?',
                              s.getvalue())
            else:
                #self.assertIn('Job up-to-date', s.getvalue())
                pass

        # one output incorrectly generated
        cleanup_tmpdir()
        with open(input_file, 'w') as outfile:
            outfile.write('testme')
        test_pipeline.run([split1], verbose=0,
                          checksum_level=CHECKSUM_HISTORY_TIMESTAMPS)
        job_history = dbdict.open(
            get_default_history_file_name(), picklevalues=True)
        del job_history[os.path.relpath(split1_outputs[0])]

        for chksm in possible_chksms:
            s = StringIO()
            test_pipeline.printout(
                s, [split1], verbose=6, checksum_level=chksm)
            if chksm >= CHECKSUM_HISTORY_TIMESTAMPS:
                self.assertIn('Job needs update:', s.getvalue())
                self.assertIn('left over from a failed run?',
                              s.getvalue())
            else:
                #self.assertIn('Job up-to-date', s.getvalue())
                pass

    def test_newstyle_merge_output(self):
        """test multiple-input checksums"""
        test_pipeline = self.create_pipeline()
        # one output incorrectly generated
        cleanup_tmpdir()
        with open(input_file, 'w') as outfile:
            outfile.write('testme')
        test_pipeline.run([split1], verbose=0,
                          checksum_level=CHECKSUM_HISTORY_TIMESTAMPS)
        job_history = dbdict.open(
            get_default_history_file_name(), picklevalues=True)
        del job_history[os.path.relpath(split1_outputs[0])]

        for chksm in possible_chksms:
            s = StringIO()
            test_pipeline.printout(
                s, [merge2], verbose=6, checksum_level=chksm)
            if chksm >= CHECKSUM_HISTORY_TIMESTAMPS:
                self.assertIn('Job needs update:', s.getvalue())
                self.assertIn('left over from a failed run?', s.getvalue())
            else:
                #self.assertIn('Job up-to-date', s.getvalue())
                pass

        # make sure the jobs run fine
        cleanup_tmpdir()
        with open(input_file, 'w') as outfile:
            outfile.write('testme')
        test_pipeline.run([merge2], verbose=0,
                          checksum_level=CHECKSUM_HISTORY_TIMESTAMPS)
        for chksm in possible_chksms:
            s = StringIO()
            test_pipeline.printout(
                s, [merge2], verbose=6, checksum_level=chksm)
            #self.assertIn('Job up-to-date', s.getvalue())
            self.assertNotIn('Job needs update:', s.getvalue())
            self.assertNotIn('left over from a failed run?', s.getvalue())

    def tearDown(self):
        shutil.rmtree(tempdir)
        pass


if __name__ == '__main__':
    unittest.main()

#        try:
#            os.mkdir(tempdir)
#        except OSError:
#            pass
#        #os.system('rm %s/*' % tempdir)
#        #open(input_file, 'w').close()
#        s = StringIO()
#        pipeline_run([transform1], checksum_level=CHECKSUM_HISTORY_TIMESTAMPS, pipeline= "main")
#        pipeline_printout(s, [transform1], verbose=6, checksum_level=0, pipeline= "main")
#        print s.getvalue()
#        #open(transform1_out)  # raise an exception if test fails



#!/usr/bin/env python
from ruffus import *
import sys

def first_task():
    print "First task"

@follows(first_task)
def second_task():
    print "Second task"

@follows(second_task)
def final_task():
    print "Final task"
pipeline_printout_graph ( "manual_follows1.png",
                         "png",
                         [final_task],
                         no_key_legend=True)

